try:
    from torch.utils.data import DataLoader, random_split
except ModuleNotFoundError:
    raise ModuleNotFoundError("Pytorch not found. Install with pip install torch")


class BaseDataLoader(DataLoader):
    def __init__(self, dataset, *args, **kwargs):
        super(BaseDataLoader, self).__init__(dataset=dataset, *args, **kwargs)

    def split(self, train: float, test: float or None = None, validation: float or None = None):
        """

        Randomly split a dataset into non-overlapping new datasets of given percentages.

        Args:
            train: percentage in train set
            test: percentage in test set (if None given: test =  1 - train)
            validation: percentage in validation set (if None: validation = 1 - train - test)

        Returns: list of DataLoader-objects for each set

        """

        if validation is None:

            train_length = int(len(self.dataset) * train)
            test_length = int(len(self.dataset) - train_length)

            return [BaseDataLoader(item) for item in random_split(self.dataset, lengths=[train_length, test_length])]

        else:
            if test is None:
                test = 1 - train - validation

            train_length = int(len(self.dataset) * train)
            test_length = int(len(self.dataset) * test)
            val_length = int(len(self.dataset) - train_length - test_length)

            return [BaseDataLoader(item) for item in
                    random_split(self.dataset, lengths=[train_length, test_length, val_length])]




class Coords:
    """ Defines geographic coordinates

    The format to use is decimal degrees (DD). For instance:

        46.9491463,7.4388499

    For the train station of Bern, the capital of Switzerland.
    """

    def __init__(self, lat: float, long: float):
        self.lat = lat
        self.long = long

    def __repr__(self):
        return "{}, {}".format(self.lat, self.long)



from .abstract_base_detector import AbstractBaseDetector
from .abstract_base_generator import AbstractBaseGenerator
from .abstract_base_model import AbstractBaseModel
from .abstract_base_time_series import AbstractBaseTimeSeries
from .abstract_io import (
    AbstractInput,
    AbstractOutputJson,
    AbstractOutputPickle,
    AbstractOutputText,
)

__all__ = [
    "AbstractBaseDetector",
    "AbstractBaseGenerator",
    "AbstractBaseModel",
    "AbstractBaseTimeSeries",
    # Abstract IO
    "AbstractInput",
    "AbstractOutputJson",
    "AbstractOutputPickle",
    "AbstractOutputText",
]



try:
    from torch.utils.data import Subset
except ModuleNotFoundError:
    raise ModuleNotFoundError("Pytorch not found. Install with pip install torch")

from .base_dataloader import BaseDataLoader


class PredictionDataLoader(BaseDataLoader):
    def __init__(self, dataset, *args, **kwargs):
        super(PredictionDataLoader, self).__init__(dataset=dataset, *args, **kwargs)

    def split(self, train: float, test: float or None = None, validation: float or None = None):
        """Split Dataset into train, test and validation

        Splitting a dataset into train, test and validation by percentages.

        Args:
            train: percentage of train split
            test: percentage of test split
            validation: percentage of validation split

        Returns: Subset of the DataSet

        """

        idx = list(range(len(self.dataset)))
        n = len(self.dataset)

        if validation is None:

            n_train = int(n * train)

            train_idx = idx[:n_train]
            test_idx = idx[n_train:]

            train_set = Subset(self.dataset, train_idx)
            test_set = Subset(self.dataset, test_idx)

            return train_set, test_set

        else:
            if test is None:
                test = 1 - train - validation

            n_train = int(n * train)
            n_test = int(n * test)

            train_idx = idx[:n_train]
            val_idx = idx[n_train:(n_train + n_test)]
            test_idx = idx[(n_train + n_test):]

            train_set = Subset(self.dataset, train_idx)
            val_set = Subset(self.dataset, val_idx)
            test_set = Subset(self.dataset, test_idx)

            return train_set, val_set, test_set



from configobj import ConfigObj
from validate import Validator
from os import path


class AnomalyConfigParser(ConfigObj):
    def __init__(self, config_file):

        super().__init__(config_file, configspec=f'{path.dirname(__file__)}/.configspec.ini')

        self.val = AnomalyConfigValidator()

        # Per default its False
        self.valid_config = False
        self.offending = []

        self._validate()

    def _validate(self):

        assert 'percent' in self['GLOBAL'] or 'selection' in self['GLOBAL'] or 'amount' in self['GLOBAL'], \
            "Either a percentage (percent), a row selection (selection) or a number of rows (amount) as to be set."

        _validator_output = self.validate(self.val)
        if _validator_output is True:
            self.valid_config = True
        else:
            self.get_offending(_validator_output)

        if self.offending:
            raise ValueError("Error in {}".format(self.offending))

    def get_offending(self, validation: dict):
        for keys, values in validation.items():
            if isinstance(values, dict):
                self.get_offending(values)
            elif isinstance(values, bool):
                if not values:
                    self.offending.append((keys, values))


class AnomalyConfigValidator(Validator):
    def __init__(self):
        super().__init__()




class Unit:
    """ Defines a physical unit of measurement, like Celsius."""

    def __init__(self, name: str, symbol: str, data_type: str):
        self.name = name
        self.symbol = symbol
        self.data_type = data_type



import pandas as pd
import numpy as np

from scipy.fft import fft, ifft

from .utils import start_indices


class AnomalyABC():
    """
        Basically just a collection of functions that can introduce anomalies into a data-series.
    """

    def __init__(self, seed: int or None = None):

        self.seed = seed
        np.random.seed(self.seed)

    def flatline(self, data: pd.Series, n: int = 1, minimum: int = 1, maximum: int or None = None):
        '''

        A sensor gets stuck at time t for a random length at the value of t-1

        Args:
            data: pandas Series of the data
            n: number of flatline events

        Returns: flatline events with coordinates where to integrate them

        '''

        if maximum is None:
            maximum = len(data)

        event_starts = start_indices(len(data), n)
        last_values = [data[e - 1] for e in event_starts]
        coordinates = []
        values = []
        for i, e in enumerate(event_starts):

            e_length = np.random.randint(minimum, maximum)
            e_max_length = data[e:e + e_length].shape[0]
            e_end = e + e_max_length
            coordinates.append([e, e_end])
            values.append(np.full((e_length), last_values[i])[:e_max_length])

        return np.array(values), coordinates

    def zeroing(self, data: pd.Series, n: int = 1, value: int = None, minimum: int = 1, maximum: int or None = None):
        '''

        A sensor stops transmitting values or falls on a default value for a random time period

        Args:
            data: pandas series with the data
            n: number of events
            value: default value (or None) the sensor falls on

        Returns: events with coordinates where to integrate them

        '''

        assert value is None or value == 0

        if maximum is None:
            maximum = len(data)

        event_starts = start_indices(len(data), n)
        coordinates = []
        values = []
        for i, e in enumerate(event_starts):

            e_length = np.random.randint(minimum, maximum)
            e_max_length = data[e:e + e_length].shape[0]
            e_end = e + e_max_length
            coordinates.append([e, e_end])
            values.append(np.full((e_length), value)[:e_max_length])

        return np.array(values), coordinates

    def outlier(self, data: pd.Series, n: int = 1, sigma: float = 0.1):
        '''

        Sudden 1 time index long spike in the time series

        Args:
            data: pandas series with the data
            n: number of events
            sigma: standard deviation of the normal distribution to draw the value to add from

        Returns: events with coordinates where to integrate them

        '''

        event_starts = start_indices(len(data), n)
        coordinates = []
        values = []
        for e in event_starts:
            offset = 1 if data[e] > 0 else -1
            values.append(data[e] + (np.random.normal(data[e], sigma, 1) * offset))
            coordinates.append([e, e])

        return np.array(values), coordinates

    def increase_noise(self, data: pd.Series, n: int = 1, sigma: float = 0.1,
            change_point_factor: float = 3.0, minimum: int = 1, maximum: int or None = None):
        '''

        # TODO: Make it possible so the noise can be decreased. (can we extract noise levels from input data)?

        A series of normal distributed values that are offset by a factor from the normal

        Args:
            data: pandas series
            n: number of events
            sigma: sigma of the normal distribution
            change_point_factor: factor * sigma for the offset of the anomaly

        Returns: events with coordinates where to integrate them

        '''

        if maximum is None:
            maximum = len(data)

        event_starts = start_indices(len(data), n)
        coordinates = []
        values = []

        for e in event_starts:
            offset = 1 if data[e] > 0 else -1
            start_value = data[e] + np.random.normal(data[e], change_point_factor * sigma,
                                                     1) * offset
            e_length = np.random.randint(minimum, maximum)
            e_max_length = data[e:e + e_length].shape[0]
            e_end = e + e_max_length

            coordinates.append([e, e_end])
            values.append(np.full((e_length), np.random.normal(start_value, sigma, e_length))[:e_max_length])

        return np.array(values), coordinates

    def change_point(self, data: pd.Series, n: int = 1, ending: bool = True,
            change_point_factor: int = 3, minimum: int = 1, maximum: int or None = None):
        '''

        A clear point t at which the data get offset to a different level

        Args:
            data: pandas time series
            n: number of events
            ending: if True the changepoint is reverted at a random time, else it goes until the end
            change_point_factor: sigma of the normal distributed change point

        Returns: events with coordinates where to integrate them

        '''

        if maximum is None:
            maximum = len(data)

        event_starts = start_indices(len(data), n)
        coordinates = []
        values = []
        end = len(data) - 1

        for e in event_starts:

            if ending:
                e_length = np.random.randint(minimum, maximum)
                end = e + e_length
            coordinates.append([e, end])
            # TODO: This give a line...introduction of noise?
            # TODO: Fixing this value might lead to issues
            offset = np.random.normal(0, change_point_factor, 1)
            # TODO: Fill this with normal distribution -> make this a user choice
            change = np.full((len(data[e:end])), offset)
            values.append(change)

        return np.array(values), coordinates

    def clipping(self, data: pd.Series, clip_value: float, mode: str = 'top'):
        """

        the time series gets clipped in the extreme values to a given user value

        Args:
            data: pandas time-series
            clip_value: value at which to cutoff.
            mode: 'top': only clip in positive, 'bottom' only in negative, 'both' do both

        Returns: events with coordinates where to integrate them

        """

        if mode == 'top':
            max_ind_top = np.where(data > clip_value)[0]
            max_ind = max_ind_top
        elif mode == 'bottom':
            max_ind_bottom = np.where(data < clip_value)[0]
            max_ind = max_ind_bottom
        else:
            invert_clip_value = clip_value * -1
            max_ind_top = np.where(data > clip_value)[0]
            max_ind_bottom = np.where(data < invert_clip_value)[0]
            max_ind = np.unique(np.concatenate((max_ind_top, max_ind_bottom), 0))

        start = np.random.choice(range(len(max_ind) + 1), 1, replace=False)[0]
        clip_ind = max_ind[start:]
        coordinates = []
        values = []
        for c in max_ind_top:
            coordinates.append([c, c])
            values.append([clip_value])
        for c in max_ind_bottom:
            coordinates.append([c, c])
            values.append([invert_clip_value])

        return np.array(values), coordinates

    def trend(self, data: pd.Series, slope: float, n: int = 1):
        '''

        the sensor as new trend in the data.

        Args:
            data: pandas series
            slope: level m of the trend according to y = m*x + b
            n: number of events

        Returns: events with coordinates where to integrate them

        '''

        event_starts = start_indices(len(data), n)
        values = []
        coordinates = []

        for start in event_starts:
            values.append(slope * np.arange(0, len(data[start:None]), 1))
            coordinates.append([start, len(data) - 1])

        return np.array(values), coordinates

    def electric_feedback(self, data: pd.Series, sampling_speed: int = 1, amp: float = 10, hrz: float = 50):
        '''

        Introducing an electric feedback into the TimeSeries using a Fast Fourier Transform.

        Args:
            data: TimeSeries object
            sampling_speed: frequency of the sampling (number of seconds between two indices)
            amp: amplitude of the feedback
            hrz: hertz of the frequency introduced

        Returns: TimeSeries object with the added frequency

        '''

        assert sampling_speed > 0

        # fft of the complete series
        N = len(data)
        #
        T = 1 / sampling_speed
        # we need to drop NaN so we can create the FFT
        yf = fft(data.dropna().to_numpy())
        xf = np.linspace(0.0, 1 / (2 * T), round(N / 2))

        # adding feedback at the given frequency (hrz)
        idx = (np.abs(xf - hrz)).argmin()
        yf[idx] += amp
        added_feedback_data = ifft(yf).real
        # extract a part of the new data and return it
        event_start = start_indices(len(yf), 1)

        # needs to return a list
        values = [added_feedback_data[event_start[0]:]]
        coordinates = [[event_start[0], len(data) - 1]]

        return np.array(values), coordinates

    def hard_knee(self, data: pd.Series, threshold: float, factor: float = 0.5, n: int = 1):

        #TODO: This might not even be a thing...it is basically a ratio_compression
        '''

        Values above a threshold are squeezed by a factor (0 < factor <= 1)

        Args:
            data: pandas series
            threshold: values above this will be squeezed
            factor: squeeze factor
            n: number of events

        Returns: events with coordinates where to integrate them

        '''

        event_starts = start_indices(len(data), n)
        values = []
        coordinates = []

        for e in event_starts:
            for i, v in enumerate(data[e:None]):
                if np.abs(v) >= threshold:
                    if v >= 0:
                        values.append([data[i] - (data[i] - threshold) * factor])
                        coordinates.append([i, i])
                    else:
                        values.append([data[i] + (data[i] + threshold) * factor])
                        coordinates.append([i, i])

        return np.array(values), coordinates

    def max_smoothing(self, data: pd.Series, window_size: int, threshold: float, n: int = 1):
        '''

        values above a threshold will be smoothed by a moving average

        Args:
            data: pandas series
            window_size: size of the moving average window -> bigger window; more smooth
            threshold: value above the sensor vales will be "smoothend"
            n: number of events

        Returns: events with coordinates where to integrate them

        '''

        event_starts = start_indices(len(data), n)
        values = []
        coordinates = []

        for e in event_starts:
            window = np.zeros(window_size)
            for i, v in enumerate(data):
                window = np.append(window, [v])
                window = np.delete(window, 0)
                avg = np.mean(window)
                if i >= e and np.abs(v) >= threshold:
                    values.append([avg])
                    coordinates.append([i, i])

        return np.array(values), coordinates

    def ratio_compression(self, data: pd.Series, threshold: float, ratio: int = 4, n: int = 1):
        '''

        values above a threshold will be compressed by a ratio

        Args:
            data: pandas series
            threshold: values above will be compressed
            ratio: factor of the compression 1:x
            n: number of events

        Returns: events with coordinates where to integrate them

        '''

        event_starts = start_indices(len(data), n)
        values = []
        coordinates = []

        for e in event_starts:
            for i, v in enumerate(data):
                if i >= e and np.abs(v) >= threshold:
                    if v >= 0:
                        values.append([data[i] - ((data[i] - threshold) / ratio)])
                        coordinates.append([i, i])
                    else:
                        values.append([data[i] - ((data[i] + threshold) / ratio)])
                        coordinates.append([i, i])

        return np.array(values), coordinates





def relative_error(x_0: float, x: float):
    """
    Compute the relative error between two values

    .. math::
        \delta x = \frac{x_0 - x}{x}

    Args:
        x_0: float of the inferred value
        x: float of the measured value

    Returns:
        float value of the relative error

    """
    return (x_0 - x) / x


def absolute_error(x_0: float, x: float):
    """
    Compute the absolute error between two values

    .. math::
        \Delta x = x_0 - x

    Args:
        x_0: float of the inferred value
        x: float of the measured value

    Returns:
        float value of the absolute error

    """
    return x_0 - x



import pandas as pd


def get_operator(mode):
    '''

    Args:
        mode: name of the mode

    Returns: mode to operate with the array

    '''

    # TODO: This could be solved more elegant with the inspect-library. For now this is good enough
    operations = {'insert': insert,
                  'replace': replace,
                  'add': add}

    return operations[mode]


def add(data, start, values):
    """

    Args:
        data: DataFrame of where we enter the new data -> only the row with "values" will be used
        start: list of start coordinates
        values: list of values to add

    Returns: Pandas Series with added anomaly

    """

    data = data['values'].values

    for e, v in zip(start, values):
        data[e[0]:e[0] + len(v)] = data[e[0]:e[0] + len(v)] + v

    return pd.Series(data)


def replace(data, start, values):
    '''

    e[0] is the starting coordinate
    e[1] is the ending coordinate

    Args:
        data: numpy array where values are inserted
        start: start index of the replacement
        values: numpy arrays with values that replace in original

    Returns:

    '''

    data = data['values'].values

    for e, v in zip(start, values):
        data[e[0]:e[0] + len(v)] = v

    return pd.Series(data)


def insert(data, start, values):
    '''

    e[0] is the starting coordinate
    e[1] is the ending coordinate

    Args:
        data: numpy array where values are inserted
        start: start index of the insertion
        values: numpy arrays to be inserted

    Returns: numpy array with values arrays inserted

    '''

    data = data['values'].values

    tmp = list(data)

    # if there are more than 1 anomalies to insert we need to check that the coordinates are corrected
    # TODO: This will fail, when the anomaly coordinates are overlapping
    if len(start) > 1:
        start_values = sorted(zip(start, values), key=lambda x: x[1][0])
        offset = 0
        for i, (coordinates, values) in enumerate(start_values):
            s, e = coordinates
            tmp = tmp[:s + offset] + list(values) + tmp[s + offset:]
            offset += e - s

    return pd.Series(tmp)



import random
from collections import defaultdict
from copy import deepcopy
from typing import List, Any, NoReturn, Tuple, Union, Optional
from warnings import warn

import numpy as np
from pandas import DataFrame, Timestamp, Timedelta, concat
from pandas.tseries.frequencies import to_offset
from pandas.tseries.offsets import DateOffset

from timeatlas.abstract import (
    AbstractBaseTimeSeries,
    AbstractOutputText,
    AbstractOutputPickle
)
from timeatlas.time_series import TimeSeries
from timeatlas.plots.time_series_dataset import line_plot
from timeatlas.utils import ensure_dir, to_pickle


class TimeSeriesDataset(List,
                        AbstractBaseTimeSeries,
                        AbstractOutputText,
                        AbstractOutputPickle):
    """ Defines a set of time series

    A TimeSeriesDataset represent a set of TimeSeries
    objects.

    """

    def __init__(self, data: List[TimeSeries] = None):
        super().__init__()
        if data is not None:
            for i in data:
                if isinstance(i, TimeSeries):
                    self.append(i)
                else:
                    raise TypeError(f"All elements of data must be a "
                                    f"TimeSeries; got {type(i)}")

    def __repr__(self):
        description = self.describe()
        return description.__repr__()

    def __setitem__(self, item, value) -> NoReturn:
        if isinstance(value, TimeSeries):
            self[item] = value
        else:
            raise ValueError("Value argument must be TimeSeries")

    # ==========================================================================
    # Methods
    # ==========================================================================

    # TimeSeries
    # ----------

    @staticmethod
    def create(length: int, start: str, end: str,
            freq: Union[str, 'TimeSeries'] = None) \
            -> 'TimeSeriesDataset':
        """
        Create an empty TimeSeriesDataset object with a defined index and period

        Args:
            length: int representing the number of TimeSeries to include in the
                TimeSeriesDataset
            start: str of the start of the DatetimeIndex
                (as in Pandas.date_range())
            end: the end of the DatetimeIndex (as in Pandas.date_range())
            freq: the optional frequency it can be a str or a TimeSeries
                (to copy its frequency)

        Returns:
            TimeSeriesDataset
        """
        # Check length parameter
        assert length >= 1, 'Length must be >= 1'
        data = []
        ts = TimeSeries.create(start, end, freq)
        for i in range(length):
            data.append(ts)
        return TimeSeriesDataset(data)

    def append(self, item: TimeSeries) -> None:
        """Append a TimeSeries to TimeSeriesDataset

        Appending an item to the TimeSeriesDataset, if item is of type TimeSeries.

        Args:
            item: item to append

        Returns: None

        """
        if not isinstance(item, TimeSeries):
            raise ValueError(f"Only items of type TimeSeries can be appended to TimeSeriesDataset; got {type(item)}")
        else:
            super().append(item)

    def plot(self) -> Any:
        """Plot a TimeSeriesDataset

        Returns:
            plotly.graph_objects.Figure
        """
        return line_plot(self)

    def copy(self, deep=True) -> 'TimeSeriesDataset':
        """Copy a list

        Copy the TSD to either a deep or shallow copy of itself

        Args:
            deep: if True, creates a deep copy else a shallow one

        Returns: (deep) copy of TimeSeriesDataset

        """
        return deepcopy(self) if deep else super().copy()

    def split_at(self, timestamp: Union[str, Timestamp]) \
            -> Tuple['TimeSeriesDataset', 'TimeSeriesDataset']:
        """Split a TimeSeriesDataset at a defined point and include the
        splitting point in both as in [start,...,at] and [at,...,end].

        Args:
            timestamp: str or Timestamp where to the TimeSeriesDataset will be
                split (e.g. "2019-12-31 00:00:00")

        Returns:
            a Tuple of TimeSeriesDataset ([start,...,at] and [at,...,end])

        """
        arr = [self[i].split_at(timestamp) for i in range(len(self))]
        first_split = TimeSeriesDataset([split_ts[0] for split_ts in arr])
        second_split = TimeSeriesDataset([split_ts[1] for split_ts in arr])
        return first_split, second_split

    def split_in_chunks(self, n: int) -> List['TimeSeriesDataset']:
        """The TimeSeries in the TimeSeriesDataset are cut into chunks
        of length n

        Args:
            n: length of the individual chunks

        Returns: List of TimeSeriesDatasets containing the chunks
        """
        # Split all TS in the TSD in chunks and store them in an array
        chunkified_ts_arr = []
        for ts in self:
            chunkified_ts = ts.split_in_chunks(n)
            chunkified_ts_arr.append(chunkified_ts)
        # Create n TSDs containing each the ith chunks of all TS. Then add each
        # TSD in an array
        chunkified_tsd_arr = []
        for ith_chunk in range(n):
            tsd_data = []
            for chunkified_ts in chunkified_ts_arr:
                tsd_data.append(chunkified_ts[ith_chunk])
            tsd = TimeSeriesDataset(tsd_data)
            chunkified_tsd_arr.append(tsd)
        return chunkified_tsd_arr

    def fill(self, value: Any) -> 'TimeSeriesDataset':
        """Fill all values in each TimeSeries from a TimeSeriesDataset.

        If given a unique value, all values will be broadcast. If given an
        array of the length of the TimeSeriesDataset, it will replace all
        values.

        Args:
            value: Any values that you want to fill the TimeSeriesDataset with

        Returns:
            TimeSeries
        """
        return TimeSeriesDataset([ts.fill(value) for ts in self])

    def empty(self) -> 'TimeSeriesDataset':
        """Empty the values in each TimeSeries from a TimeSeriesDataset.

        This method is equivalent to filling all values of each TimeSeries in a
        TimeSeriesDataset with NaNs.

        Returns:
            TimeSeriesDataset
        """
        return self.fill(np.nan)

    def pad(self, limit: Union[int, str, Timestamp], side: Optional[str] = None,
            value: Any = np.NaN) -> 'TimeSeriesDataset':
        """
        Pad a TimeSeriesDataset until a given limit

        Args:
            limit: int, str or Pandas Timestamp
                if int, it will pad the side given in the side arguments by n
                elements.

            side: Optional[str]
                side to which the TimeSeries will be padded. This arg can have
                two value: "before" and "after" depending where the padding is
                needed.

                This arg is needed only in case the limit is given in int.

            value: Any values

        Returns:
            TimeSeriesDataset
        """
        return TimeSeriesDataset([ts.pad(limit=limit, side=side, value=value)
                                  for ts in self])

    def trim(self, side: str = "both") -> 'TimeSeriesDataset':
        """Remove NaNs from a TimeSeries start, end or both

        Args:
            side:
                the side where to remove the NaNs. Valid values are either
                "start", "end" or "both". Default to "both"

        Returns:
            TimeSeries
        """
        return TimeSeriesDataset([ts.trim(side) for ts in self])

    def merge(self, tsd: 'TimeSeriesDataset') -> 'TimeSeriesDataset':
        """Merge two TimeSeriesDataset by the index of the TimeSeries

        This methods merges the TimeSeries from the TimeSeriesDataset (TSD) in
        argument with self based on the indexes of each one of the TSDs.

        Args:
            tsd: the TimeSeriesDataset to merge with self

        Returns:
            TimeSeriesDataset
        """
        arr = []
        for i, ts in enumerate(self):
            merged_ts = ts.merge(tsd[i])
            arr.append(merged_ts)
        return TimeSeriesDataset(arr)

    def merge_by_label(self, tsd: 'TimeSeriesDataset') -> 'TimeSeriesDataset':
        """Merge two TimeSeriesDatasets by the label of the TimeSeries
        in the TimeSeriesDatasets

        Args:
            tsd: TimeSeriesDataset to be merged with self

        Returns:
            TimeSeriesDataset
        """

        def list_duplicates(seq: list) -> list:
            """Get label duplicates

            Getting a dict of with key=ts.label and value=index in TSD

            Args:
                seq: list of labels

            Returns: dict of duplicate indices

            """
            tally = defaultdict(list)
            for i, item in enumerate(seq):
                tally[item].append(i)

            duplicates_items = ((key, locs) for key, locs in tally.items())

            duplicates = []
            for dup in sorted(duplicates_items):
                duplicates.append(dup)

            return duplicates

        def merge_duplicates(tsd: TimeSeriesDataset, duplicates: list) \
                -> TimeSeriesDataset:
            """Merging based on duplicates

            Merging the TimeSeriesDataset based on the duplicate list.

            Args:
                tsd: TimeSeriesDataset to be merged
                duplicates: list of tuples with (label,duplicate indices)

            Returns: merged TimeSeriesDataset

            """
            arr = []
            for label, inds in duplicates:
                # We merge everything into the first occurrence of the label
                # -> tsd[inds[0]]
                base_ts = tsd[inds[0]]
                if len(inds) > 1:
                    for ind in inds[1:]:
                        base_ts = base_ts.merge(tsd[ind])
                    arr.append(base_ts)
                else:
                    arr.append(base_ts)

            return TimeSeriesDataset(arr)

        # First : Create new TSD and add all elements
        merged_tsd = self.copy()
        for ts in tsd:
            merged_tsd.append(ts)

        # Second : Get list of tuples (label, duplicate indices)
        tsd_labels = [ts.label for ts in merged_tsd]
        duplicates = list_duplicates(seq=tsd_labels)

        # Third : Merge TSD, where TS has same labels.
        merged_tsd = merge_duplicates(tsd=merged_tsd, duplicates=duplicates)

        return merged_tsd

    # TimeSeriesDataset
    # -----------------

    def select_components_randomly(self, n: int, seed: int = None,
            indices: bool = False) -> Any:
        """Returns a subset of the TimeSeriesDataset with randomly chosen n
        elements without replacement.

        Args:
            n: number of elements returned
            seed: seed for random generator
            indices: if True returns the indices of the selection

        Returns: TimeSeriesDataset (optional: indices of selection)

        """
        # setting the seed if None no seed will be set automatically
        random.seed(seed)
        if indices:
            inds, data = zip(*random.sample(
                population=list(enumerate(self)), k=n))
            return list(inds), TimeSeriesDataset(list(data))
        else:
            TimeSeriesDataset(random.sample(population=self, k=n))

    def select_components_by_percentage(self, percent: float, seed: int = None,
            indices: bool = False) -> Any:
        """Returns a subset of the TimeSeriesDataset with randomly chosen
        percentage elements without replacement.

        Args:
            percent: percentage of elements returned
            seed: seed for random generator
            indices: if True returns the indices of the selection

        Returns: TimeSeriesDataset (optional: indices of selection)

        """
        # setting the seed if None no seed will be set automatically
        random.seed(seed)
        n = round(len(self) * percent)

        # Workaround: If percentage too small we select at least 1
        if n <= 0:
            warn(
                f'set percentage to small resulting selection is <= 0\n Using n=1.')
            n = 1
        if indices:
            return self.select_components_randomly(n=n, indices=indices)
        else:
            return self.select_components_randomly(n=n)

    def shuffle(self, inplace: bool = False) -> 'TimeSeriesDataset':
        """Randomizing the order of the TS in the TSD

        Randomizing the order of the TS in TSDs and returning a new TSD.

        Args:
            inplace: randomizing inplace or creating new object. (Default: False)

        Returns: shuffled TimeSeriesDataset

        """

        if inplace:
            random.shuffle(self)
        else:
            new_tsd = self.copy()
            random.shuffle(new_tsd)
            return new_tsd

    # ==========================================================================
    # Processing
    # ==========================================================================

    # TimeSeries
    # ----------

    def apply(self, func, tsd: 'TimeSeriesDataset' = None) \
            -> 'TimeSeriesDataset':
        # TODO (See GitHub issue 56)
        raise NotImplementedError

    def resample(self, freq: Union[str, TimeSeries],
            method: Optional[str] = None) \
            -> 'TimeSeriesDataset':
        """Convert the TimeSeries in a TimeSeriesDataset to a specified
        frequency. Optionally provide filling method to pad/backfill missing
        values.

        Args:
            freq: str or TimeSeries.
                The new time difference between two adjacent entries in the
                returned TimeSeries.

                If a TimeSeries is given, the freq in self will become
                the same as in the given TimeSeries.

                If a string is given, three options are available:
                    * 'lowest' : to sync the TimeSeriesDataset to the lowest frequency
                    * 'highest' : to sync the TimeSeriesDataset to the highest frequency
                    * A DateOffset alias e.g.: 15min, H, etc.


            method: {'backfill'/'bfill', 'pad'/'ffill'}, default None
                Method to use for filling holes in reindexed Series (note this
                does not fill NaNs that already were present):

                * 'pad'/'ffil': propagate last valid observation forward to next
                  valid
                * 'backfill'/'ffill': use next valid observation to fill

        Returns:
            TimeSeriesDataset
        """
        # Acting differently depending on freq arg type and value
        if isinstance(freq, TimeSeries):
            target_freq = freq.frequency()
        elif freq == "lowest":
            frequencies = [to_offset(ts.frequency()) for ts in self]
            target_freq = max(frequencies)
        elif freq == "highest":
            frequencies = [to_offset(ts.frequency()) for ts in self]
            target_freq = min(frequencies)
        elif isinstance(to_offset(freq), DateOffset):
            target_freq = freq
        else:
            raise ValueError("freq argument isn't valid")
        # Resample and return
        return TimeSeriesDataset(
            [ts.resample(target_freq, method) for ts in self])

    def group_by(self, freq: str, method: Optional[str] = "mean") \
            -> 'TimeSeriesDataset':
        """Groups values by a frequency for each TimeSeries in a
        TimeSeriesDataset.

        This method is quite similar to resample with the difference that it
        gives the guaranty that the timestamps are full values.
        e.g. 2019-01-01 08:00:00.

        Resample could make values spaced by 1 min but
        every x sec e.g. [2019-01-01 08:00:33, 2019-01-01 08:01:33],
        which isn't convenient for further index merging operations.

        The function has different aggregations methods taken from Pandas
        groupby aggregations[1]. By default, it'll take the mean of the
        defined freq bucket.

        [1] https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#aggregation

        Args:
            freq: string offset alias of a frequency
            method: string of the Pandas aggregation function.

        Returns:
            TimeSeriesDataset
        """
        return TimeSeriesDataset([ts.group_by(freq, method) for ts in self])

    def interpolate(self, *args, **kwargs) -> 'TimeSeriesDataset':
        """Wrapper around the Pandas interpolate() method.

        See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.interpolate.html
        for reference
        """
        return TimeSeriesDataset(
            [ts.interpolate(*args, **kwargs) for ts in self])

    def normalize(self, method: str) -> 'TimeSeriesDataset':
        """Normalize the TimeSeries in a TimeSeriesDataset with a given method

        Args:
            method: str
            * 'minmax' for a min max normalization
            * 'zscore' for Z score normalization

        Returns:
            TimeSeriesDataset
        """
        return TimeSeriesDataset([ts.normalize(method) for ts in self])

    def round(self, decimals: int) -> 'TimeSeriesDataset':
        """Round the values of every TimeSeries in the TimeSeriesDataset with a
        defined number of digits

        Args:
            decimals: int defining the number of digits after the comma

        Returns:
            TimeSeriesDataset
        """
        return TimeSeriesDataset([ts.round(decimals) for ts in self])

    def sort(self, *args, **kwargs) -> 'TimeSeriesDataset':
        """Sort the TimeSeries of a TimeSeriesDataset by time stamps

        Basically, it's a wrapper around df.sort_index()
        see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.sort_index.html

        Args:
            args: the positional arguments
            kwargs: the keyword arguments

        Returns:
            TimeSeriesDataset
        """
        return TimeSeriesDataset([ts.sort(*args, **kwargs) for ts in self])

    # TimeSeriesDataset
    # -----------------
    def regularize(self, side: str = "[]", fill=np.nan):
        """
        Regularize a TimeSeriesDataset so that all starting and ending
        timestamps are similar. This method keeps the frequency of every
        TimeSeries the same.

        fill : user defined value, mean, median, etc.
        method : ][, [[, ]], []

        return
            TSD
        """
        # 1. find the boundaries of all TSs
        all_boundaries = self.boundaries()

        # 2. find earliest/latest values
        starts = np.array(all_boundaries).T[0]
        earliest_start = min(starts)
        latest_start = max(starts)
        ends = np.array(all_boundaries).T[1]
        earliest_end = min(ends)
        latest_end = max(ends)

        # 3. pad all TSs accordingly
        if side == "[]":
            tsd = self.pad(earliest_start).pad(latest_end)
        elif side == "[[":
            tsd = TimeSeriesDataset([ts[:earliest_end]
                                     for ts in self.pad(earliest_start)])
        elif side == "]]":
            tsd = TimeSeriesDataset([ts[latest_start:]
                                     for ts in self.pad(latest_end)])
        elif side == "][":
            tsd = TimeSeriesDataset([ts[latest_start:earliest_end]
                                     for ts in self])
        else:
            raise ValueError("method argument is not recognized")

        return tsd

    # ==========================================================================
    # Analysis
    # ==========================================================================

    # Basic Statistics
    # ----------------

    def min(self) -> List[Any]:
        return [ts.min() for ts in self]

    def max(self) -> List[Any]:
        return [ts.max() for ts in self]

    def mean(self) -> List[Any]:
        return [ts.mean() for ts in self]

    def median(self) -> List[Any]:
        return [ts.median() for ts in self]

    def kurtosis(self) -> List[Any]:
        return [ts.kurtosis() for ts in self]

    def skewness(self) -> List[Any]:
        return [ts.skewness() for ts in self]

    def describe(self) -> DataFrame:
        """Describe a TimeSeriesDataset with the describe function from Pandas

        TODO Define what describe should do on TimeSeriesDataset (see issue 56)

        Returns:
            TODO Define return type
        """
        min = self.min()
        max = self.max()
        mean = self.mean()
        median = self.median()
        kurtosis = self.kurtosis()
        skewness = self.skewness()

        return DataFrame.from_dict({'minimum': min,
                                    'maximum': max,
                                    'mean': mean,
                                    'median': median,
                                    'kurtosis': kurtosis,
                                    'skewness': skewness})

    # Time Series Statistics
    # ----------------------

    def start(self) -> List[Timestamp]:
        """Get the first Timestamp of a all components of
        a TimeSeriesDataset

        Returns:
            List of Pandas Timestamp
        """
        return [ts.start() for ts in self]

    def end(self) -> List[Timestamp]:
        """Get the last Timestamp of a all components of
        a TimeSeriesDataset

        Returns:
            List of Pandas Timestamp
        """
        return [ts.end() for ts in self]

    def boundaries(self) -> List[Tuple[Timestamp, Timestamp]]:
        """Get the tuple with the TimeSeries first and last index for all
        components in the TimeSeriesDataset

        Returns:
            List of Tuple of Pandas Timestamps
        """
        return [ts.boundaries() for ts in self]

    def frequency(self) -> List[Optional[str]]:
        """Get the frequency of a each TimeSeries in a TimeSeriesDataset

        Returns:
            List of str or None
                - str of the frequency according to the Pandas Offset Aliases
                - None if no discernible frequency
        """
        return [ts.frequency() for ts in self]

    def resolution(self) -> 'TimeSeriesDataset':
        """Compute the time difference between each timestamp for all TimeSeries
        in a TimeSeriesDataset

       Returns:
           TimeSeriesDataset
       """
        return TimeSeriesDataset(
            [ts.resolution() for ts in self])

    def duration(self) -> List[Timedelta]:
        """Get the duration for all TimeSeries in a TimeSeriesDataset

        Returns:
            a List of Pandas Timedelta
        """
        return [ts.duration() for ts in self]

    # =============================================
    # IO
    # =============================================

    # Outputs

    def to_text(self, path: str) -> NoReturn:
        """Export a TimeSeriesDataset to text format

        Args:
            path: Path, where the TimeSeriesDataset will be saved in

        Returns: NoReturn
        """
        ensure_dir(path)
        for i, ts in enumerate(self):
            ts_path = "{}/{}".format(path, i)
            ensure_dir(ts_path)
            ts.to_text(ts_path)

    def to_pickle(self, path: str) -> NoReturn:
        """Creating a pickle out of the TimeSeriesDataset

        Args:
            path: Path, where the TimeSeriesDataset will be saved

        Returns: NoReturn
        """
        to_pickle(self, path)

    def to_df(self) -> DataFrame:
        """Converts a TimeSeriesDataset to a Pandas DataFrame

        The indexes of all the TimeSeries in the TimeSeriesDataset will get
        merged. That means that you are the only responsible if the merge
        induces the addition of many NaNs in your data. Therefore, it's better
        to use this method if you are sure that your TimeSeries share a common
        frequency as well as start and end.

        Returns:
            DataFrame
        """

        # get the result ready
        columns = []
        res = self[0].to_df()
        for i, col in enumerate(res.columns):
            columns.append(f"{col}_{i}")
        # merge the rest in the result
        rest = self[1:]
        for i, ts in enumerate(rest):
            res = res.merge(ts.series, how="outer",
                            left_index=True, right_index=True)
            cols = ts.series.columns
            for col in cols:
                columns.append(f"{col}_{i+1}")
        res.columns = columns
        return res

    def to_array(self) -> np.ndarray:
        """TimeSeriesData to NumpyArray [n x len(tsd)], where n is number of TimeSeries in dataset
        # TODO Should output a warning if the ts have different length
            (issue 56)

        Returns: numpy.array of shape (n x len(tsd))
        """
        return np.array([ts.to_array() for ts in self], dtype=object)

    def to_darts(self):
        """Convert a TimeSeriesDataset to Darts TimeSeries

        Returns:
            Darts TimeSeries object
        """
        # TODO issue 56
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import NoReturn


class AbstractBaseGenerator(ABC):
    """ Abstract class to define methods to implement
    for a Generator class.
    """

    @abstractmethod
    def __init__(self):
        # Each generator should be called the super class and set a label_suffix
        self.label_suffix = None

    @abstractmethod
    def generate(self) -> NoReturn:
        """ Generate features """
        raise NotImplementedError



import numpy as np

import inspect


def start_indices(maximum: int, n: int):
    '''

    e[0] is the starting coordinate
    e[1] is the ending coordinate

    Args:
        maximum: max integer to draw a number from
        n: how many numbers to draw

    Returns: starting index of the events to be inserted

    '''
    return np.random.randint(0, maximum, n)


def get_function_names(anomaly_object: object):
    """

    Args:
        anomaly_object: object containing the functions

    Returns: a list of functions excl. __init__

    """
    functions = [func[0] for func in inspect.getmembers(anomaly_object, predicate=inspect.ismethod) if
                 not func[0].startswith('__')]

    return functions



from typing import NoReturn, Tuple, Any, Union, Optional, List, Callable

from darts import TimeSeries as DartsTimeSeries
import numpy as np
from pandas import DataFrame, date_range, infer_freq, Series, DatetimeIndex, \
    Timestamp, Timedelta

from timeatlas.abstract import (
    AbstractBaseTimeSeries,
    AbstractOutputText,
    AbstractOutputPickle
)
from timeatlas.config.constants import (
    TIME_SERIES_VALUES,
    TIME_SERIES_FILENAME,
    TIME_SERIES_EXT,
    METADATA_FILENAME,
    METADATA_EXT
)
from timeatlas.metadata import Metadata
from timeatlas.processors.scaler import Scaler
from timeatlas.plots.time_series import line_plot, status_plot
from timeatlas.utils import ensure_dir, to_pickle


class TimeSeries(AbstractBaseTimeSeries, AbstractOutputText,
                 AbstractOutputPickle):
    """Defines a time series

    A TimeSeries object is a series of time indexed values.

    Attributes:
        series: An optional Pandas DataFrame
        metadata: An optional Dict storing metadata about this TimeSeries
    """

    def __init__(self, series: Union[Series, DataFrame] = None,
                 metadata: Metadata = None, label: str or None = None):

        if series is not None:
            # Check if values have a DatetimeIndex
            assert isinstance(series.index, DatetimeIndex), \
                'Values must be indexed with a DatetimeIndex.'

            # Check if the length is greater than one
            assert len(series) >= 1, 'Values must have at least one values.'

            # Save the data with the right format
            if isinstance(series, Series):
                series.name = TIME_SERIES_VALUES
                series = series.to_frame()

            elif isinstance(series, DataFrame):
                assert len(series.columns) >= 1, \
                    "DataFrame as input series must have at least one column."

                # If there's only one column, then it is the values column
                if len(series.columns) == 1:
                    series.columns = [TIME_SERIES_VALUES]

                # Otherwise, one column should be called "values"
                assert TIME_SERIES_VALUES in series.columns, \
                    "DataFrame as input series must contain a column called {}"\
                    .format(TIME_SERIES_VALUES)

            # Create the TimeSeries object with certainty that values
            # are sorted by the time index
            self.series = series.sort_index()

            # Add the freq if regular
            if len(series) >= 3:
                infer_freq(self.series.index)

            # Create instance variables
            self.index = self.series.index  # index accessor
            self.values = self.series[TIME_SERIES_VALUES]  # values accessor
        else:
            self.series = None

        self.label = label  # label of the TimeSeries (for classification)

        if metadata is not None:
            self.metadata = metadata
        else:
            self.metadata = None

    def __repr__(self):
        return self.series.__repr__()

    def __len__(self):
        return len(self.series)

    def __iter__(self):
        return (v for v in self.series[TIME_SERIES_VALUES])

    def __getitem__(self, item):
        return TimeSeries(self.series[item])

    def __setitem__(self, item, value):
        if isinstance(value, TimeSeries):
            self.series[item] = value.series[item]
        else:
            self.series[item] = value[item]

    # ==========================================================================
    # Methods
    # ==========================================================================

    @staticmethod
    def create(start: str, end: str, freq: Union[str, 'TimeSeries'] = None,
               metadata: Metadata = None) \
            -> 'TimeSeries':
        """Creates an empty TimeSeries object with the period as index

        Args:
            start: str of the start of the DatetimeIndex (as in Pandas.date_range())
            end: the end of the DatetimeIndex (as in Pandas.date_range())
            freq: the optional frequency it can be a str or a TimeSeries (to copy its frequency)
            metadata: the optional Metadata object

        Returns:
            TimeSeries
        """
        if freq is not None:
            if isinstance(freq, TimeSeries):
                freq = infer_freq(freq.series.index)
            elif isinstance(freq, str):
                freq = freq
        series = DataFrame(columns=[TIME_SERIES_VALUES],
                           index=date_range(start, end, freq=freq))
        return TimeSeries(series, metadata)

    def plot(self) -> Any:
        """Plot a TimeSeriesDataset

        Returns:
            plotly.graph_objects.Figure
        """
        return line_plot(self)

    def split_at(self, timestamp: Union[str, Timestamp]) \
            -> Tuple['TimeSeries', 'TimeSeries']:
        """Split a TimeSeries at a defined point and include the splitting point
        in both as in [start,...,at] and [at,...,end].

        Args:
            timestamp: str or Timestamp where to the TimeSeries will be split
                            (e.g. "2019-12-31 00:00:00")

        Returns:
            a Tuple of TimeSeries ([start,...,at] and [at,...,end])

        """
        start = self.series.index[0]
        end = self.series.index[-1]
        first_split = self.series[start:timestamp].copy()
        second_split = self.series[timestamp:end].copy()
        before = TimeSeries(first_split, self.metadata)
        after = TimeSeries(second_split, self.metadata)
        return before, after

    def split_in_chunks(self, n: int) -> List['TimeSeries']:
        """Split a TimeSeries into chunks of length n

        Args:
            n: length of the chunks

        Returns:
            List of TimeSeries
        """
        ts_chunks = [TimeSeries(series=v, metadata=self.metadata) for n, v in
                     self.series.groupby(np.arange(len(self.series)) // n)]
        return ts_chunks

    def fill(self, value: Any) -> 'TimeSeries':
        """Fill a TimeSeries with a value. If given a unique value, all values
        will be broadcasted. If given an array of the length of the TimeSeries,
        it will replace all values.

        Args:
            value: Any values that you want to fill the TimeSeries with

        Returns:
            TimeSeries
        """
        s = self.series.copy()
        s[TIME_SERIES_VALUES] = value
        return TimeSeries(s, self.metadata)

    def empty(self) -> 'TimeSeries':
        """Empty the TimeSeries (fill all values with NaNs)

        Returns:
            TimeSeries
        """
        return self.fill(np.nan)

    def pad(self, limit: Union[int, str, Timestamp], side: Optional[str] = None,
            value: Any = np.NaN):
        """
        Pad a TimeSeries until a given limit

        Args:
            limit: int, str or Pandas Timestamp
                if int, it will pad the side given in the side arguments by n
                elements.

            side: Optional[str]
                side to which the TimeSeries will be padded. This arg can have
                two value: "before" and "after" depending where the padding is
                needed.

                This arg is needed only in case the limit is given in int.

            value: Any values

        Returns:
            TimeSeries
        """
        def create_pad(new_limit, ts, fill_val):
            """
            Local utility function to create padding time series from a Pandas
            Timestamp for a given TimeSeries

            Args:
                new_limit: Pandas Timestamp of the new limit to pad from/to
                ts: TimeSeries to pad
                fill_val: value to fill the TimeSeries with

            Returns:
                TimeSeries
            """
            if new_limit < ts.start():
                return ts.create(new_limit, ts.start(), freq=ts)\
                           .fill(fill_val)[:-1]
            elif new_limit > ts.end():
                return ts.create(ts.end(), new_limit, freq=ts)\
                           .fill(fill_val)[1:]
            if new_limit == ts.start() or new_limit == ts.end():
                return TimeSeries()
            else:
                raise ValueError("The given limit is included in the time "
                           "series, padding is impossible")

        # Create padding TimeSeries from a given number of elements to pad with
        if isinstance(limit, int):
            # Add 1 to make the interval is too small
            target_limit = limit + 1

            if side == "before":
                index = date_range(end=self.start(), freq=self.frequency(),
                                   periods=target_limit, closed="left")
            elif side == "after":
                index = date_range(start=self.end(), freq=self.frequency(),
                                   periods=target_limit, closed="right")
            else:
                raise ValueError("side argument isn't valid")

            values = [value] * len(index)
            df = DataFrame(index=index, data=values)
            pad = TimeSeries(df)

        # Create padding TimeSeries from time stamp as str
        if isinstance(limit, str):
            target_limit = Timestamp(limit)
            pad = create_pad(target_limit, self, value)

        # Create padding TimeSeries from time stamp as Pandas Timestamp
        elif isinstance(limit, Timestamp):
            target_limit = limit
            pad = create_pad(target_limit, self, value)

        else:
            ValueError("limit argument isn't valid")

        return self.merge(pad)

    def trim(self, side: str = "both") -> 'TimeSeries':
        """Remove NaNs from a TimeSeries start, end or both

        Args:
            side:
                the side where to remove the NaNs. Valid values are either
                "start", "end" or "both". Default to "both"

        Returns:
            TimeSeries
        """
        if side == "both":
            first = self.series.first_valid_index()
            last = self.series.last_valid_index()
            series_wo_nans = self.series[TIME_SERIES_VALUES].loc[first:last]
        elif side == "start":
            first = self.series.first_valid_index()
            series_wo_nans = self.series[TIME_SERIES_VALUES].loc[first:]
        elif side == "end":
            last = self.series.last_valid_index()
            series_wo_nans = self.series[TIME_SERIES_VALUES].loc[:last]
        else:
            raise AttributeError("side attribute must be either 'start' or "
                                 "'end', but not {}".format(side))
        return TimeSeries(series_wo_nans, self.metadata)

    def merge(self, ts: 'TimeSeries') -> 'TimeSeries':
        """Merge two time series and make sure all the given indexes are sorted.

        Args:
            ts: the TimeSeries to merge with self

        Returns:
            TimeSeries
        """
        # append and infer new freq
        merged = self.series.append(ts.series)
        infer_freq(merged.index)

        # instanciate a TimeSeries to sort it
        return TimeSeries(merged, self.metadata)

    # ==========================================================================
    # Processing
    # ==========================================================================

    def apply(self, func, ts: 'TimeSeries' = None) -> 'TimeSeries':
        """Wrapper around the Pandas apply function

        Args:
            func: function
                Python function or NumPy ufunc to apply. If ts is given as
                param, func must have two params in the form of f(x,y)

            ts: TimeSeries
                The second TimeSeries if you want to make an operation on two
                time series

        Returns:
            TimeSeries
        """
        if ts is not None:
            s1 = self.series[TIME_SERIES_VALUES]
            s2 = ts.series[TIME_SERIES_VALUES]
            df = DataFrame(data={"s1": s1, "s2": s2})
            res = TimeSeries(df.apply(lambda x: func(x.s1, x.s2), axis=1),
                             self.metadata)
        else:
            res = TimeSeries(self.series[TIME_SERIES_VALUES].apply(func),
                             self.metadata)

        return res

    def resample(self, freq: str, method: Optional[str] = None) -> 'TimeSeries':
        """Convert TimeSeries to specified frequency. Optionally provide filling
        method to pad/backfill missing values.

        Args:
            freq: string
                The new time difference between two adjacent entries in the
                returned TimeSeries. A DateOffset alias is expected.

            method: {'backfill'/'bfill', 'pad'/'ffill'}, default None
                Method to use for filling holes in reindexed Series (note this
                does not fill NaNs that already were present):

                * 'pad'/'ffil': propagate last valid observation forward to next
                  valid
                * 'backfill'/'ffill': use next valid observation to fill

        Returns:
            TimeSeries
        """
        # avoid duplicated indexes
        new_series = self.series[~self.series.index.duplicated()]
        new_series = new_series.asfreq(freq, method=method)
        return TimeSeries(new_series, self.metadata)

    def group_by(self, freq: str, method: Optional[str] = "mean")\
            -> 'TimeSeries':
        """Groups values by a frequency.

        This method is quite similar to resample with the difference that it
        gives the guaranty that the timestamps are full values.
        e.g. 2019-01-01 08:00:00.

        Resample could make values spaced by 1 min but
        every x sec e.g. [2019-01-01 08:00:33, 2019-01-01 08:01:33],
        which isn't convenient for further index merging operations.

        The function has different aggregations methods taken from Pandas
        groupby aggregations[1]. By default, it'll take the mean of the
        defined freq bucket.

        [1] https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#aggregation

        Args:
            freq: string offset alias of a frequency
            method: string of the Pandas aggregation function.

        Returns:
            TimeSeries
        """
        # Group by freq
        series = self.series.groupby(self.index.round(freq))

        # Apply aggregation
        if method == "mean":
            series = series.mean()
        elif method == "sum":
            series = series.sum()
        elif method == "size":
            series = series.size()
        elif method == "count":
            series = series.count()
        elif method == "std":
            series = series.std()
        elif method == "var":
            series = series.var()
        elif method == "sem":
            series = series.sem()
        elif method == "first":
            series = series.first()
        elif method == "last":
            series = series.last()
        elif method == "min":
            series = series.min()
        elif method == "max":
            series = series.max()
        else:
            ValueError("method argument not recognized.")

        return TimeSeries(series, self.metadata)

    def interpolate(self, *args, **kwargs) -> 'TimeSeries':
        """Wrapper around the Pandas interpolate() method.

        See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.interpolate.html
        for reference
        """
        new_series = self.series.interpolate(*args, **kwargs)
        return TimeSeries(new_series, self.metadata)

    def normalize(self, method: str) -> 'TimeSeries':
        """Normalize a Time Series with a given method

        Args:
            method: str
            * 'minmax' for a min max normalization
            * 'zscore' for Z score normalization

        Returns:
            TimeSeries
        """
        if method == "minmax":
            scaled_series = Scaler.minmax(self.series)
        elif method == "zscore":
            scaled_series = Scaler.zscore(self.series)
        else:
            raise ValueError("{} isn't recognized as a normalization method"
                             .format(method))
        return TimeSeries(scaled_series, self.metadata)

    def round(self, decimals: int) -> 'TimeSeries':
        """Round the values in the series.values

        Args:
            decimals: number of digits after the comma
        """
        return TimeSeries(self.series.astype(float).round(decimals=decimals),
                          metadata=self.metadata)

    def sort(self, *args, **kwargs):
        """Sort a TimeSeries by time stamps

        Basically, it's a wrapper around df.sort_index()
        see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.sort_index.html

        Args:
            args: the positional arguments
            kwargs: the keyword arguments

        Returns:
            TimeSeries
        """
        sorted_time_series = self.series.sort_index(*args, **kwargs)
        return TimeSeries(sorted_time_series, self.metadata)

    # ==========================================================================
    # Analysis
    # ==========================================================================

    # Basic Statistics
    # ----------------

    def min(self) -> float:
        """Get the minimum value of a TimeSeries

        Returns:
            float
        """
        return self.series[TIME_SERIES_VALUES].min()

    def max(self) -> float:
        """Get the maximum value of a TimeSeries

        Returns:
            float
        """
        return self.series[TIME_SERIES_VALUES].max()

    def mean(self) -> float:
        """Get the mean value of a TimeSeries

        Returs:
            float
        """
        return self.series[TIME_SERIES_VALUES].mean()

    def median(self) -> float:
        """Get the median value of a TimeSeries

        Returns:
            float
        """
        return self.series[TIME_SERIES_VALUES].median()

    def skewness(self) -> float:
        """Get the skewness of a TimeSeries

        Returns:
            float
        """
        return self.series[TIME_SERIES_VALUES].skew()

    def kurtosis(self) -> float:
        """Get the kurtosis of a TimeSeries

        Returns:
            float
        """
        return self.series[TIME_SERIES_VALUES].kurtosis()

    def describe(self, percentiles=None, include=None, exclude=None) -> Series:
        """Describe a TimeSeries with the describe function from Pandas

        Returns:
            Series
        """
        return self.series[TIME_SERIES_VALUES].describe()

    # Time Series Statistics
    # ----------------------

    def start(self) -> Timestamp:
        """Get the first Timestamp of a TimeSeries

        Returns:
            Timestamp
        """
        start = self.series.index[0]
        return start

    def end(self) -> Timestamp:
        """Get the last Timestamp of a TimeSeries

        Returns:
            Timestamp
        """
        end = self.series.index[-1]
        return end

    def boundaries(self) -> Tuple[Timestamp, Timestamp]:
        """Get a tuple with the TimeSeries first and last index

        Returns:
            a Tuple of Pandas Timestamps
        """
        start = self.series.index[0]
        end = self.series.index[-1]
        return start, end

    def frequency(self) -> Optional[str]:
        """Get the frequency of a TimeSeries

        Returns:
            str or None
            str of the frequency according to the Pandas Offset Aliases
            (https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
            None if no discernible frequency
        """
        return self.series.index.inferred_freq

    def resolution(self) -> 'TimeSeries':
        """Compute the time difference between each timestamp of a TimeSeries

        Returns:
            TimeSeries
        """
        diff = self.series.index.to_series().diff().dt.total_seconds()
        ts = TimeSeries(DataFrame(diff, columns=[TIME_SERIES_VALUES]),
                        self.metadata)
        return ts

    def duration(self) -> Timedelta:
        """Get the duration of the TimeSeries

        Returns:
            a Pandas Timedelta
        """
        start, end = self.boundaries()
        return end - start

    # ==========================================================================
    # IO
    # ==========================================================================

    def to_text(self, path: str) -> NoReturn:
        """Export a TimeSeries to text format

        Args:
            path: str of the path where the TimeSeries will be saved
        """
        # Create the time series file
        file_path = "{}/{}.{}".format(path, TIME_SERIES_FILENAME,
                                      TIME_SERIES_EXT)
        ensure_dir(file_path)
        self.__series_to_csv(self.series, file_path)
        if self.metadata is None and self.label is not None:
            self.metadata = Metadata(items={'label': self.label})
        # Create the metadata file
        if self.metadata is not None:
            if self.label is not None:
                self.metadata.add({'label': self.label})
            file_path = "{}/{}.{}".format(path, METADATA_FILENAME, METADATA_EXT)
            ensure_dir(file_path)
            self.metadata.to_json(pretty_print=True, path=file_path)

    def to_array(self) -> np.ndarray:
        """Convert a TimeSeries to Numpy Array

        Returns:
            numpy.array with dimensions (1, n), where n is th length of the
            TimeSeries
        """
        return self.series[TIME_SERIES_VALUES].to_numpy()

    def to_pickle(self, path: str) -> NoReturn:
        """Export a TimeSeries to Pickle

        Args:
            path: str of the path where the TimeSeries will be saved
        """
        to_pickle(self, path)

    def to_darts(self) -> DartsTimeSeries:
        """Convert a TimeSeries to Darts TimeSeries

        Returns:
            Darts TimeSeries object
        """
        return DartsTimeSeries.from_series(self.series[TIME_SERIES_VALUES])

    def to_df(self) -> DataFrame:
        """Converts a TimeSeries to a Pandas DataFrame

        Returns:
            Pandas DataFrame
        """
        return self.series

    @staticmethod
    def __series_to_csv(series: Union[Series, DataFrame], path: str) \
            -> NoReturn:
        """Export a Pandas Series or DataFrame and put it into a CSV file

        Args:
            series: Pandas Series or Pandas DataFrame
                The data to write in CSV
            path: str
                The path where the Series will be saved
        """
        series.to_csv(path, header=True, index_label="index")



from abc import ABC, abstractmethod
from typing import NoReturn


class AbstractBaseDetector(ABC):
    """ Abstract class to define methods to implement
    for a Detector class.
    """

    @abstractmethod
    def detect(self, ts) -> NoReturn:
        """ Detect features """
        raise NotImplementedError



import pickle
from typing import Any


def read_pickle(path: str) -> Any:
    """
    Load an object from a pickle file

    Args:
        path: str of the path of the object on your file system

    Returns:
       The deserialized object
    """
    with open(path, 'rb') as f:
        return pickle.load(f)



from matplotlib.axes import Axes


def add_metadata_to_plot(meta, ax: Axes):
    if meta is not None:
        if "unit" in meta:
            unit = meta["unit"]
            ax.set_ylabel("{} $[{}]$".format(unit.name, unit.symbol))
        if "sensor" in meta:
            sensor = meta["sensor"]
            ax.set_title("{}—{}".format(sensor.id, sensor.name))
    return ax



from typing import TYPE_CHECKING, Any
from datetime import datetime

import numpy as np
import seaborn as sns
from matplotlib import (
    pyplot as plt,
    dates as mdates
)
from pandas.plotting import register_matplotlib_converters
import plotly.graph_objects as go
import plotly.io as pio


from timeatlas.config.colors import colors
from ._utils import add_metadata_to_plot
from timeatlas.config.constants import (
    TIME_SERIES_VALUES,
    TIME_SERIES_CI_LOWER,
    TIME_SERIES_CI_UPPER,
)

if TYPE_CHECKING:
    from timeatlas.time_series import TimeSeries

pio.templates.default = "plotly_white"
sns.set_style("whitegrid")
sns.set_context("notebook")


def line_plot(ts: 'TimeSeries', context: str = "notebook", *args, **kwargs)\
        -> Any:
    """
    Plot a TimeSeries

    This is a wrapper around Pandas.Series.plot() augmented if the
    TimeSeries to plot has associated Metadata.

    Args:
        ts: the TimeSeries to plot
        context: str defining on which medium the plot will be displayed
        *args: positional arguments for Pandas plot() method
        **kwargs: keyword arguments fot Pandas plot() method

    Returns:
        matplotlib.axes._subplots.AxesSubplot
    """
    if context == "paper":
        register_matplotlib_converters()

        if 'figsize' not in kwargs:
            kwargs['figsize'] = (18, 2)  # Default TimeSeries plot format

        if 'color' not in kwargs:
            kwargs['color'] = colors.blue_dark

        ax = ts.series.plot(*args, **kwargs)
        ax.set_xlabel("Date")
        ax.grid(True, c=colors.grey, ls='-', lw=1, alpha=0.2)

        # Add legend from metadata if existing
        if ts.metadata is not None:
            if "unit" in ts.metadata:
                unit = ts.metadata["unit"]
                ax.set_ylabel("{} $[{}]$".format(unit.name, unit.symbol))
            if "sensor" in ts.metadata:
                sensor = ts.metadata["sensor"]
                ax.set_title("{}—{}".format(sensor.id, sensor.name))

        return ax
    elif context == "notebook":
        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x=ts.index,
            y=ts.values,
            name=ts.series[TIME_SERIES_VALUES].name
        ))
        fig.update_layout(showlegend=True)
        fig.show()
        return fig
    else:
        raise ValueError("Context doesn't exit")


def prediction_plot(forecast: 'TimeSeries', observation: 'TimeSeries' = None) -> Any:
    """
    Make a plot to display a time series with forecasted values and its
    confidence interval. If given, the observation time series can be added as
    an optional argument.

    Args:
        forecast: TimeSeries of the prediction
        observation: TimeSeries of the historical data

    Returns:
        matplotlib.axes._subplots.AxesSubplot
    """
    fig = plt.figure(figsize=(18, 4))
    ax = fig.add_subplot(111)
    ax.grid(True, which='major', c=colors.blue_dark, ls='-', lw=1, alpha=0.2)
    ax.set_xlabel("Date")

    # Add contextual information to the plot
    if forecast.metadata is not None:
        if "sensor" in forecast.metadata:
            sensor = forecast.metadata["sensor"]
            duration = str(forecast.duration())
            sensor_str = "{}—{}".format(sensor.id, sensor.name)
            ax.set_title("Forecast of {} for {}".format(sensor_str, duration))
        else:
            ax.set_title("Forecast for {}".format(str(forecast.duration())))

        if "unit" in forecast.metadata:
            unit = forecast.metadata["unit"]
            ax.set_ylabel("{} $[{}]$".format(unit.name, unit.symbol))

    # Add the lines to the plot
    ax.plot(forecast.series.index, forecast.series[TIME_SERIES_VALUES].values,
            ls='--',
            c=colors.blue_dark,
            label="prediction")

    # If present, add confidence interval
    if TIME_SERIES_CI_LOWER in forecast.series.columns and \
       TIME_SERIES_CI_UPPER in forecast.series.columns:
        ax.fill_between(forecast.series.index,
                        forecast.series[TIME_SERIES_CI_LOWER].values,
                        forecast.series[TIME_SERIES_CI_UPPER].values,
                        color=colors.blue_light,
                        label="confidence interval")

    if observation is not None:
        ax.plot(observation.series.index, observation.series.values,
                ls='-',
                c=colors.blue_dark,
                label="observation")

    ax.legend()
    return ax


def status_plot(ts: 'TimeSeries', cmap: str = "autumn_r") -> Any:
    """
    Plot a uni-dimensional imshow to mimic status plots like on
    https://githubstatus.com

    Args:
        ts: TimeSeries - the time series to plot
        cmap: String - the matplotlib colormap to use (default is "automn_r")

    Returns:
        matplotlib.axes._subplots.AxesSubplot
    """
    register_matplotlib_converters()

    fig, ax = plt.subplots(figsize=(18,1))

    # Set x limits
    x_lims = [ts.boundaries()[0].to_pydatetime(),ts.boundaries()[1].to_pydatetime()]
    x_lims = mdates.date2num(x_lims)

    # Set y limits (for the sake of having something...)
    y_lims = [ts.min().values[0], ts.max().values[0]]

    date_format = mdates.DateFormatter('%d/%m/%y %H:%M:%S')
    ax.xaxis.set_major_formatter(date_format)

    ax.set_yticks([])  # remove all yticks
    ax.xaxis_date()

    m = ax.imshow([ts.series],
              extent=[x_lims[0], x_lims[1],  y_lims[0], y_lims[1]],
              aspect='auto',
              cmap=cmap)

    ax = add_metadata_to_plot(ts.metadata, ax)

    fig.autofmt_xdate()
    plt.grid(b=True, which='both')
    plt.colorbar(m, aspect=5, pad=0.01)
    plt.show()
    return ax


def kde_plot(ts: 'TimeSeries') -> np.ndarray:
    """
    Display a KDE plot through time with a line plot underneath

    Args:
        ts: TimeSeries - the time series to plot

    Returns:
        numpy.ndarray[matplotlib.axes._subplots.AxesSubplot]
    """
    timestamps = ts.series.index.astype(np.int64) // 10 ** 9
    values = ts.series["values"].values

    fig, axs = plt.subplots(2, 1,
                            sharex='col',
                            figsize=(16, 5),
                            gridspec_kw={'height_ratios': [4, 1]})

    upper_plot = axs[0]
    lower_plot = axs[1]

    sns.kdeplot(timestamps, values,
                shade=True,
                cmap="YlOrRd",
                shade_lowest=False,
                ax=upper_plot)

    x_ticks = upper_plot.get_xticks()
    x_labels = [datetime.fromtimestamp(t).strftime('%Y-%m-%d %H:%M') for t in x_ticks]

    upper_plot = add_metadata_to_plot(ts.metadata, upper_plot)
    upper_plot.set_xticklabels(x_labels)
    upper_plot.xaxis.tick_bottom()

    lower_plot.plot(timestamps, values, c=colors.blue_dark)
    lower_plot.xaxis.tick_top()
    lower_plot.grid()
    return axs



import numpy as np


def chunkify(tsd, seq_len):
    """

    Splits a TimeSeriesDataset into chunks of length seq_len

    Args:
        tsd: TimeSeriesDataset object
        seq_len: length of the subsequences to return

    Returns: numpy arrays with chunks of size seq_len

    """
    x, y = [], []
    for s in tsd:
        for i in range(len(s) - seq_len):
            x_i = s.series["values"][i: i + seq_len]
            y_i = s.series["values"][i + seq_len]

            x.append(x_i)
            y.append(y_i)

    return np.array(x), np.array(y)



from typing import NoReturn, Tuple, Any, Union, Optional, List, Callable, Dict

from timeatlas.abstract.abstract_base_generator import AbstractBaseGenerator
from timeatlas.time_series import TimeSeries
from timeatlas.time_series_dataset import TimeSeriesDataset
from timeatlas.config.constants import TIME_SERIES_VALUES

from .anomalies import AnomalyABC
from .utils import get_operator
from .labeler import AnomalySetLabeler
from .config import AnomalyConfigParser

import pandas as pd
import numpy as np
from itertools import cycle
from copy import copy
import math
from os import path


class AnomalyGenerator(AbstractBaseGenerator):
    """

    A generator that introcudes an anomaly into a given TimeSeriesDataset.

    The types and parameters are controlled with a .ini file,
    that can be created with "AnomalyGeneratorTemplate"

    """

    def __init__(self, data: TimeSeriesDataset, conf_file: str, save_as: str = 'text'):
        """

        Args:
            data: TimeSeriesDataset containing the data
            conf_file: config file created with AnomalyGeneratorTemplate
        """

        # Each generator set a label_suffix
        # Here: AGM -> Anomaly Generator Manual
        super().__init__()
        self.label_suffix = "AGM"
        self.save_as = save_as

        # assertions
        assert isinstance(data, TimeSeriesDataset)
        assert all(isinstance(x, TimeSeries) for x in
                   data), "One or more elements are not a TimeSeries-object"
        assert path.isfile(
            conf_file), f"No config file found under given path '{conf_file}'"

        # set data
        self.data = data

        # read the config file
        self.config = AnomalyConfigParser(config_file=conf_file)
        self.GLOBAL = self.config['GLOBAL']
        self.ANOMALIES = self.config['ANOMALIES']
        self.selection = self.GLOBAL['selection']
        self.percent = self.GLOBAL['percent']
        self.amount = self.GLOBAL['amount']
        self.outfile = self.GLOBAL['outfile']

        # create numpy-random.RandomState object
        self.seed = self.GLOBAL['seed']

        # functions for anomaly
        self.ABC = AnomalyABC(self.seed)
        self.anomaly_functions = self.get_anomaly_function()

        # adding a label column to the dataframe and creating the results anomaly labels
        self.labels = AnomalySetLabeler()

        # figure out the precision of the data
        self.precision = self.generation_precision()

    @staticmethod
    def precision_and_scale(x: float):
        """

        Get the precision of a value

        Args:
            x: a (float) number

        Returns: the number of positions after the comma

        """
        # 14 is the maximal number of digits python can handle (more is also unrealistic)
        max_digits = 14
        # if the number is NaN return nothing
        if math.isnan(x):
            return
        # figure out the magniture -> the numbers before the comma
        int_part = int(abs(x))
        magnitude = 1 if int_part == 0 else int(math.log10(int_part)) + 1
        if magnitude >= max_digits:
            return (magnitude, 0)
        # shift the number after the comma in front of the comma and figure out the amount
        frac_part = abs(x) - int_part
        multiplier = 10 ** (max_digits - magnitude)
        frac_digits = multiplier + int(multiplier * frac_part + 0.5)
        while frac_digits % 10 == 0:
            frac_digits /= 10
        scale = int(math.log10(frac_digits))
        return scale

    @staticmethod
    def clean_parameters(values) -> Dict:
        """
        Function to cleanup the parameters. If the parameter in the config-file are None, they are removed.
        Args:
            values: parameter values from he config files

        Returns: Dict of the paramters without the None

        """
        return {k: v for k, v in values['PARAMETERS'].items() if v is not None}

    @staticmethod
    def create_zip_object(data, anomaly_f):
        '''

        combines the two lists of the data, where the anomalies are added to and the anomaly-function

        if the function list is shorter it will cycle through them until all data has 1 anomaly

        if the data is shorter it will only assign one anomaly function

        Args:
            data: pd.Series of data
            anomaly_f: function of ABC.anomalies creating the anomaly

        Returns: zip-object

        '''

        # warnings.warn("Length of data > length of anomalies: Not all anomalies will be assigned.")

        zip_list = zip(data, cycle(anomaly_f))
        return zip_list

    def generation_precision(self):
        '''

        Set the rounded average precision of the values inside a dataframe

        Returns: rounded average number of digits after the comma

        '''

        precision_df = np.array(
            [self.precision_and_scale(x) for ts in self.data for x in
             ts.series.values])
        # This is more of a security. A correctly formated TimeSeries-object has no None elements
        precision_df = precision_df[precision_df != None]

        return int(round(precision_df.mean()))

    def save(self) -> NoReturn:
        """

        Saving the labels and the new TimeSeriesDataset to file.

        Returns: NoReturn

        """

        self.labels.finalize()

        if self.save_as == 'text':
            self.data.to_text(path=f'./{self.outfile}_data')
        elif self.save_as == 'pickle':
            self.data.to_pickle(path=f'./{self.outfile}_data.pkl')

        # This function is no longer needed, since we save the labels now in the TimeSeries
        # self.labels.annotation.to_csv(f'./{self.outfile}_data/{self.outfile}_labels.csv', index=False)

    def get_anomaly_function(self) -> List:
        '''

        Get all functions in the config file

        Returns: list of tuples with the functions as (function, parameters)

        '''
        functions = []
        for key, values in self.ANOMALIES.items():
            function = getattr(self.ABC, values['function'])
            # removing the keys with None
            parameters = self.clean_parameters(values)
            functions.append((function, parameters))
        return functions

    def chose_amount(self) -> List:
        """

        Chose the number of time windows based on a fixed amount given by the user in the config file:

        eg. amount = 10, will select 10 elements

        Returns: List of pair of indices and data

        """

        ind, data = self.data.random(n=self.amount, seed=self.seed, indices=True)
        return list(zip(ind, data))

    def chose_selection(self) -> List:
        """

        Chose the number of time windows based on a user selection given by the user in the config file:

        eg. selection = [0,1,5,9] will select the first, second, sixth and tenth element.

        Returns: List of pair of indices and data

        """
        ind, data = self.data.select(selection=self.selection, indices=True)
        return list(zip(ind, data))

    def chose_percentage(self) -> List:
        """

        Chose the number of time windows based on a user selection given by the user in the config file:

        e.g. percent = 0.2 will select 20% of the TimeSeriesDataset (min=0, max=1)

        Returns: List of pair of indices and data

        """
        ind, data = self.data.percent(percent=self.percent, seed=self.seed, indices=True)
        return list(zip(ind, data))

    def add_data(self, new_data: TimeSeries, index: int) -> NoReturn:
        """

        Replacing the old TimeSeries with the new TimeSeries containing the anomaly.

        Args:
            new_data: new TimeSeries that will replace the old one
            index: index of the TimeSeries to replace in the TimeSeriesDataset

        Returns: NoReturn

        """

        self.data[index].series[TIME_SERIES_VALUES].replace(to_replace=pd.Series(new_data))

    def add_labels(self, index, coordinates, function_name):
        """

        Create the labels that need to be added to the TimeSeries.
        Will create a new column for the labels and name them.

        Args:
            index: index of the TimeSeries in the TimeSeriesDataframe
            coordinates: start and end index of the anomaly in the TimeSeries
            function_name: label of the anomaly

        Returns:

        """
        labels = [None] * len(self.data[index].series)
        for coords in coordinates:
            start = coords[0]
            end = coords[1] + 1
            labels[start:end] = [function_name] * len(labels[start:end])
            self.data[index].series[f'label_{self.label_suffix}'] = labels
            self.data[index].label = function_name

    def generate(self) -> NoReturn:
        """
        raise NotImplementedError

        Main function to generate the anomalies.

        Returns: NoReturn

        """

        if self.amount:
            anomaly_series = self.chose_amount()
        elif self.selection:
            anomaly_series = self.chose_selection()
        else:
            anomaly_series = self.chose_percentage()

        zip_list_functions = self.create_zip_object(anomaly_series,
                                                    self.anomaly_functions)

        # TODO: This adds the anomalies at the start and not where they belong
        for (ind, ts), (function, params) in zip_list_functions:
            data = ts.series
            operation_param = params['operation']
            function_params = copy(params)
            function_params.pop('operation')
            # TODO: Here we make DataFrame -> Series. A more elegant solution is to be found
            anomaly, coordinates = function(data[TIME_SERIES_VALUES], **function_params)
            # creating the new data to add
            operator = get_operator(mode=operation_param)
            new_data = operator(data, start=coordinates, values=anomaly)
            # rounding the data to a precision typical for the given dataset
            new_data = new_data.round(decimals=self.precision)
            self.add_data(new_data=new_data, index=ind)
            self.labels.create_operation_dict(coordinates=coordinates,
                                              param=operation_param,
                                              function_name=function.__name__,
                                              name=ind,
                                              outfile=self.outfile)

            self.add_labels(index=ind,
                            coordinates=coordinates,
                            function_name=function.__name__)

        if self.GLOBAL['save']:
            self.save()



import json
from typing import Dict, Any
from timeatlas.types import *
from timeatlas.utils import ensure_dir


class Metadata(dict):

    def __init__(self, items: Dict[str, Any] = None):
        super().__init__()
        # If the object is known, transform them!
        if items is not None:
            self.add(items)

    def add(self, items: Dict[str, Any]):
        for k, v in items.items():
            if k == 'sensor':
                self[k] = v if isinstance(v, Sensor) else Sensor(v['sensor_id'], v['name'])
            elif k == 'unit':
                self[k] = v if isinstance(v, Unit) else Unit(v['name'], v['symbol'], v['data_type'])
            else:
                self[k] = v

    def to_json(self, pretty_print=False, path: str = None):
        """ Convert the current Metadata object into a JSON string

        Args:
            pretty_print: Boolean allowing for pretty printing
            path: String of the JSON file to write

        Returns:
            A String containing the JSON

        """
        my_json = json.dumps(self,
                             default=lambda x: x.__dict__,
                             sort_keys=True,
                             indent=2) \
            if pretty_print \
            else json.dumps(self,
                            default=lambda x: x.__dict__,
                            sort_keys=True)

        if path is not None:
            ensure_dir(path)
            with open(path, 'w', encoding='utf-8') as file:
                file.write(my_json)

        return my_json



""" TimeAtlas API Definition """

from timeatlas import abstract
from timeatlas import detectors
from timeatlas import generators
from timeatlas.metadata import Metadata
from timeatlas import metrics
from timeatlas import plots
from timeatlas import processors
from timeatlas.read_write import (
    read_pickle,
    read_text,
    read_tsd,
    csv_to_tsd
)
from timeatlas.time_series import TimeSeries
from timeatlas.time_series_dataset import TimeSeriesDataset
from timeatlas import types



from timeatlas.time_series_dataset import TimeSeriesDataset
from timeatlas.models.NN.util import chunkify
from .base_dataset import BaseDataset

class TimeSeriesPredictionDataset(BaseDataset):
    """
    A DataLoader for the prediction of a TimeSeries n next steps, where
    X: TimeSeries n previous steps
    y:  next step of the TimeSeries
    """

    def __init__(self, timeseriesdataset: TimeSeriesDataset, n: int or None):
        """

        Args:
            data: TimeSeriesDataset
            n: number of previous steps
        """
        super(TimeSeriesPredictionDataset, self).__init__(tsd=timeseriesdataset)
        self.data, self.labels = chunkify(tsd=timeseriesdataset.data, seq_len=n)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item], self.labels[item]



from typing import NoReturn

from configobj import ConfigObj
import inspect

from ..anomalies import AnomalyABC
from ..anomalies.utils import get_function_names


class AnomalyGeneratorTemplate(ConfigObj):
    def __init__(self, filename, seed: int = None, functions: list or str = "__all__",
            threshold: float = None, num_anomalies: int = None,
            anomaly_name: str = "ANOMALY"):
        super().__init__()

        # setting up the seed
        if seed is None:
            self.initial_comment.append("WARNING: No seed was set. This will make the results not reproducible")

        self.seed = seed

        # creating the anomaly object
        self.ABC = AnomalyABC(seed=self.seed)

        # check functions and number of anomalies
        if functions and num_anomalies:
            assert len(functions) == num_anomalies

        # setting up the list of anomalies and their functions
        if functions:
            # check if functions are defined as list, str or "__all__"
            if isinstance(functions, list):
                self.functions = functions
            elif isinstance(functions, str):
                if functions == '__all__':
                    # if all functions get them from the anomaly object
                    self.functions = get_function_names(self.ABC)
                else:
                    self.functions = [functions]

            self.num_anomalies = len(self.functions)

        # if only number of anomalies are given set up empty confic
        elif num_anomalies:
            self.num_anomalies = num_anomalies
        else:
            raise Exception("Either num_anomalies(int) or functions(list) has to be defined")

        # setup filename and comments
        self.filename = filename + '.ini'
        self.initial_comment = ["Automatically created config-file "]

        # some internal parameters by ConfigObj
        self.write_empty_values = True

        # set header name and anomaly_name
        self.header_name = 'GLOBAL'
        self.anomaly_name = anomaly_name + ' '

        self.threshold = threshold

        self.create_config()

    def create_config(self) -> NoReturn:
        """

        Creating the config file based on the input of the user

        Returns: NoReturn -> saves the file directly.

        """
        self[self.header_name] = {}
        self.inline_comments[
            self.header_name] = "!!One of the settings 'percent', 'selection' or 'amount' has to be set!!"
        self[self.header_name]['on_threshold'] = ''
        self[self.header_name]['seed'] = '' if self.seed is None else self.seed
        self[self.header_name]['percent'] = ''
        self[self.header_name]['selection'] = ''
        self[self.header_name]['amount'] = ''
        self[self.header_name]['outfile'] = ''
        self[self.header_name]['save'] = True

        self['ANOMALIES'] = {}
        self.inline_comments['ANOMALIES'] = "'operation' is either 'add', 'replace' or 'insert'"

        for i in range(0, self.num_anomalies, 1):
            n = i + 1
            self['ANOMALIES'][self.anomaly_name + str(n)] = {}
            self['ANOMALIES'][self.anomaly_name + str(n)]['PARAMETERS'] = {}
            if self.functions is not None:
                self['ANOMALIES'][self.anomaly_name + str(n)]['function'] = self.functions[i]
                params = self.anomaly_function_parameters(i)
                for param in params:
                    self['ANOMALIES'][self.anomaly_name + str(n)]['PARAMETERS'][param] = ''
            else:
                self['ANOMALIES'][self.anomaly_name + str(n)]['function'] = ''

    def anomaly_function_parameters(self, function_index: int) -> list:
        """

        Get the function parameters form the function so they can be written into the config.ini

        Args:
            function_index: index of the anomaly to the the paramters from

        Returns: List of parameters

        """
        try:
            params = inspect.getfullargspec(getattr(self.ABC, self.functions[function_index])).args
            # removing the two parameters self and data, that should not be given in the config file
            params.remove("self")
            params.remove("data")
            # adding the needed "operation" to the parameters
            params.insert(0, "operation")
            return params
        except Exception as e:
            fs = [bound_method[0] for bound_method in inspect.getmembers(self.ABC, predicate=(inspect.ismethod))]
            raise Exception("Declared function-name unknown: options: {}".format(str(fs))) from e



from typing import List


def ceil(ts: 'TimeSeries', thresholds: List) -> 'TimeSeries':
    """
    Ceil (TODO change name. proposal: stepper, threshold_overtaking, ...) is a
    function providing a time series informing the user if a threshold has been
    trespassed with natural numbers. For instance:

    0 = no trespassing,
    1 = first level trespassed,
    2 = second level trespassed, etc...

    :param ts: TimeSeries on which the data will be analyzed
    :param thresholds: List of threshold
    :return: TimeSeries of threshold trespassing
    """
    def threshold(value: float, levels: List):
        for k, v in enumerate(levels):
            if value > v:
                if k+1 == len(levels):
                    # meaning there's no other levels...
                    return k+1
                else:
                    continue
            else:
                return k

    return ts.apply(lambda x: threshold(x, thresholds))



from typing import List

try:
    import torch
    import torch.nn as nn
    from torch.utils.data import DataLoader
    from torch.optim.optimizer import Optimizer
except ModuleNotFoundError:
    raise ModuleNotFoundError("Pytorch not found. Install with pip install torch")

from timeatlas.abstract import AbstractBaseModel


class LSTMPrediction(nn.Module, AbstractBaseModel):

    def __init__(self, n_features, n_hidden, seq_len, optimizer: Optimizer, loss_function, n_layers=1,
            bias=True, batch_first=False, dropout=0, bidirectional=False, horizon: int = 1, **kwargs):
        """

        Args:
            model: object instance of a LSTM from timeatlas (e.g. model=Prediction(<parameters>)
            epochs: number of epochs on the training set
            optimizer: class of the optimizer -> see Pytroch -> torch.optim
            loss_function: a function that calculates the loss -> see PyTorch -> torch.nn
            n_features: Number of features as input
            n_hidden: number of cells in each hidden layer
            seq_len: length of each featurs
            n_layers: number of hidden layer
            bias: If False, then the layer does not use bias weights b_ih and b_hh.
            batch_first: If True, then the input and output tensors are provided as (batch, seq, feature).
            dropout: If non-zero, introduces a Dropout layer on the outputs of each LSTM layer except the last layer, with dropout probability equal to dropout.
            bidirectional: if True, becomes a bidirectional LSTM.
            horizon: the number of returned values from the Linear output layer
            **kwargs: kwargs for the optimizer
        """

        super(LSTMPrediction, self).__init__()

        self.n_hidden = n_hidden
        self.seq_len = seq_len
        self.n_layers = n_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.out_features = horizon

        self.lstm = nn.LSTM(input_size=n_features,
                            hidden_size=n_hidden,
                            num_layers=n_layers,
                            bias=bias,
                            batch_first=batch_first,
                            dropout=dropout,
                            bidirectional=bidirectional
                            )

        self.linear = nn.Linear(in_features=n_hidden, out_features=self.out_features)
        self.reset_hidden_state()

        # optimizer and loss function for the fit
        self.optimizer = optimizer(self.parameters(), **kwargs)
        self.loss_function = loss_function

        # saving the history
        self.loss = []

    def reset_hidden_state(self):
        self.hidden = (
            torch.zeros(self.n_layers, self.seq_len, self.n_hidden).double(),
            torch.zeros(self.n_layers, self.seq_len, self.n_hidden).double()
        )

    def forward(self, sequences):
        lstm_out, self.hidden = self.lstm(sequences.view(len(sequences), self.seq_len, -1),
                                          self.hidden)

        last_time_step = lstm_out.view(self.seq_len, len(sequences), self.n_hidden)[-1]
        y_pred = self.linear(last_time_step)
        return y_pred

    def fit(self, series: DataLoader, epochs: int = 10, verbose: bool = True):
        super().fit(series)
        for i in range(epochs):
            for seq, labels in self.X_train:

                self.reset_hidden_state()

                self.optimizer.zero_grad()

                y_pred = self(seq)

                single_loss = self.loss_function(y_pred.squeeze(), labels)
                single_loss.backward()
                self.optimizer.step()

                self.loss.append(single_loss.item())

        if verbose:
            print(f'epoch: {i:3} loss: {single_loss.item():10.8f}')

    def predict(self, horizon: int = 1, testset: DataLoader = None) -> List:
        super().predict(horizon)

        if testset is None:
            raise Exception(f'the testset has to be of type {type(DataLoader)}, got {type(testset)}')

        self.eval()
        predictions = []
        for seq, l in testset:
            y_pred = self(seq)
            predictions.append(y_pred)

        return predictions



# Generic
DEFAULT_EXPORT_FILENAME = "export"

# Metadata
METADATA_FILENAME = "meta"
METADATA_EXT = "json"
METADATA_CLASS_LABEL = "label"

# TimeSeries
TIME_SERIES_VALUES = "values"
TIME_SERIES_CI_LOWER = "ci_lower"
TIME_SERIES_CI_UPPER = "ci_upper"
TIME_SERIES_FILENAME = "data"
TIME_SERIES_EXT = "csv"
TIME_SERIES_LABEL_PREFIX = "label_"

# Model
MODEL_TYPE_UNIVARIATE = "univariate"
MODEL_TYPE_MULTIVARIATE = "multivariate"

# Misc. IO
PICKLE_EXT = "pkl"



from abc import ABC, abstractmethod
from typing import Any, NoReturn


class AbstractInput(ABC):
    """ Definition of method signatures for
    data input into time atlas objects
    """

    @abstractmethod
    def read(self, path: str) -> Any:
        raise NotImplementedError


class AbstractOutputText(ABC):
    """ Definition of method signatures for
    data output in text format from time atlas objects
    """

    @abstractmethod
    def to_text(self, path: str) -> NoReturn:
        raise NotImplementedError


class AbstractOutputPickle(ABC):
    """ Definition of method signatures for
    data output in Pickle format from time atlas objects
    """

    @abstractmethod
    def to_pickle(self, path: str) -> NoReturn:
        raise NotImplementedError


class AbstractOutputJson(ABC):
    """ Definition of method signatures for
        data output in JSON format from time atlas objects
    """

    @abstractmethod
    def to_json(self, path: str) -> NoReturn:
        raise NotImplementedError



from typing import NoReturn, Union, Tuple
from numpy import ndarray
from pandas import Series
from sklearn import linear_model

from timeatlas.time_series import TimeSeries
from timeatlas.abstract import AbstractBaseModel
from timeatlas.config.constants import TIME_SERIES_VALUES


class LinearRegression(AbstractBaseModel):

    def __init__(self):
        super().__init__()
        self.model = linear_model.LinearRegression()

    def fit(self, series: TimeSeries) -> NoReturn:
        """
        Fit a linear regression model given a time series

        Args:
            series: the TimeSeries to fit
        """
        super().fit(series)
        X_train, y_train = self.__prepare_series_for_sklearn(self.X_train)
        self.model.fit(X_train, y_train)

    def predict(self, horizon: Union[str, TimeSeries], freq: str = None) \
            -> TimeSeries:
        """
        Predict a TimeSeries given a horizon

        Args:
            horizon: str as in https://pandas.pydata.org/pandas-docs/stable/user_guide/timedeltas.html
            freq: frequency in DateOffset string https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

        Returns:
            TimeSeries
        """
        super().predict(horizon)
        if isinstance(horizon, str):
            future, index = self.make_future_arrays(horizon)
        elif isinstance(horizon, TimeSeries):
            future, y_train = self.__prepare_series_for_sklearn(horizon)
            index = horizon.series.index
        forecast = self.model.predict(future)
        return TimeSeries(Series(data=forecast, index=index))

    @staticmethod
    def __prepare_series_for_sklearn(ts: TimeSeries)\
            -> Tuple[ndarray, ndarray]:
        """
        Prepare a TimeSeries object so that it can be given to a Scikit Learn
        model for prediction or other task

        Args:
            ts: the TimeSeries to prepare

        Returns:
            A Tuple with two elements:
                - a Numpy ndarray with shape (n, 1)
                - a Numpy ndarray with shape (n,)
        """
        X_train = ts.series.index.factorize()[0].reshape(-1, 1)
        y_train = ts.series[TIME_SERIES_VALUES].to_numpy()
        return X_train, y_train

    def make_future_arrays(self, horizon: str, freq: str = None)\
            -> Tuple[ndarray, ndarray]:
        """
        Create the arrays needed for the prediction

        Args:
            horizon: str as in https://pandas.pydata.org/pandas-docs/stable/user_guide/timedeltas.html
            freq: frequency in DateOffset string https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

        Returns:
            A Tuple with two elements:
                - a Numpy ndarray with shape (n, 1)
                - a Numpy ndarray with shape (n,)

        """
        index = self.make_future_index(horizon, freq)
        X_train = self.__prepare_series_for_sklearn(self.X_train)
        X_test = index.factorize()[0].reshape(-1, 1) + len(X_train)
        return X_test, index




import pandas as pd
from glob import glob
from warnings import warn

from typing import Any

from timeatlas.config.constants import *
from timeatlas.time_series import TimeSeries
from timeatlas.time_series_dataset import TimeSeriesDataset
from timeatlas.config.constants import METADATA_CLASS_LABEL
from ._utils import (
    check_directory_structure,
    csv_to_dataframe,
    json_to_metadata,
)


def read_text(path: str) -> Any:
    """
    Create a TimeSeries object from a text file

    Args:
        path: a String representing the path to the file to read

    Returns:
       TimeSeries

    """

    dir_type = check_directory_structure(path)

    if dir_type == "timeseries":
        data = "{}/{}.{}".format(path, TIME_SERIES_FILENAME, TIME_SERIES_EXT)
        series = csv_to_dataframe(data)
        return TimeSeries(series)

    elif dir_type == "timeseries with metadata":
        data = "{}/{}.{}".format(path, TIME_SERIES_FILENAME, TIME_SERIES_EXT)
        meta = "{}/{}.{}".format(path, METADATA_FILENAME, METADATA_EXT)
        series = csv_to_dataframe(data)
        metadata = json_to_metadata(meta)
        ts = TimeSeries(series, metadata)
        if METADATA_CLASS_LABEL in ts.metadata:
            ts.label = ts.metadata[METADATA_CLASS_LABEL]
        return ts

    elif dir_type is None:
        raise IOError("The path doesn't' include any recognizable files")


def read_tsd(path: str) -> TimeSeriesDataset:
    """

    Loading the data from the individual data "csv"-files. These files should be in the format created by
    the "to_text"-function, either by TimeSeries.to_text() or TimeSeriesDataset.to_text().

    Exp:

    data
    │    ├── TSD-folder
    │    │    ├── 0
    │    │    │   └── data.csv
    │    │    ├── 1
    │    │    │   └── data.csv
    │    │    ├── 2
    │    │    │   └── data.csv
    │    │    ├── 3
    │    │    │   └── data.csv
    │    │    ├── 4
    │    │    │   └── (if there is no data the folder will be skipped with a warning)
    │    │    └── 4
    │    │        └── data.csv

    Args:
        path: Path to the folder that contains the subfolder containing the individual data csv-files

    Returns:
        TimeSeriesDataset

    """
    ts_list = []

    folders = glob(f'{path}/*')
    for f in folders:
        try:
            ts_list.append(read_text(f))
        except IOError:
            warn(f'Folder "{f}" does not contain any recognizable files')

    return TimeSeriesDataset(ts_list)


def csv_to_tsd(path: str) -> 'TimeSeriesDataset':
    """

    Create a TimeSeriesDataset from a csv

    Args:
        path: the path to the csv file

    Returns: TimeSeriesDataset

    """
    tsd = []
    df = pd.read_csv(path, index_col=0)
    df.index = pd.to_datetime(df.index)
    for i, d in df.iteritems():
        tsd.append(TimeSeries(d))

    return TimeSeriesDataset(data=tsd)



from typing import TYPE_CHECKING, Any

import plotly.graph_objects as go
import plotly.io as pio

if TYPE_CHECKING:
    from timeatlas.time_series_dataset import TimeSeriesDataset

pio.templates.default = "plotly_white"


def line_plot(tsd: 'TimeSeriesDataset') -> Any:
    """Plot a TimeSeriesDataset with Plotly

    Args:
        tsd: TimeSeriesDataset

    Returns:
        plotly.graph_objects.Figure
    """
    fig = go.Figure()
    for ts in tsd:
        fig.add_trace(
            go.Scatter(x=ts.index,
                       y=ts.values,
                       mode='lines'))
    return fig



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union, Tuple, List, Callable, Optional, \
    NoReturn
from pandas import DataFrame, Timestamp, Timedelta

if TYPE_CHECKING:
    from timeatlas.time_series import TimeSeries
    from timeatlas.time_series_dataset import TimeSeriesDataset
    from timeatlas.abstract.abstract_base_model import AbstractBaseModel


class AbstractBaseTimeSeries(ABC):
    """ Definition of the methods signatures
    usable for Time Series
    """

    # ==========================================================================
    # Methods
    # ==========================================================================

    @abstractmethod
    def create(self, *args) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def plot(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    @abstractmethod
    def split_at(self, timestamp: Union[str, Timestamp]) \
            -> Union[Tuple['TimeSeries', 'TimeSeries'],
                     Tuple['TimeSeriesDataset', 'TimeSeriesDataset']]:
        raise NotImplementedError

    @abstractmethod
    def split_in_chunks(self, n: int) \
            -> Union[List['TimeSeries'], List['TimeSeriesDataset']]:
        raise NotImplementedError

    @abstractmethod
    def fill(self, value: Any) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def empty(self) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def pad(self, *args) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def trim(self, side: str) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def merge(self, ts: Union['TimeSeries', 'TimeSeriesDataset']) \
            -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    # ==========================================================================
    # Processing
    # ==========================================================================

    @abstractmethod
    def apply(self, func, ts: Union['TimeSeries', 'TimeSeriesDataset'] = None) \
            -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def resample(self, freq: Union[str],
                 method: Optional[Union[str, 'AbstractBaseModel']] = None) \
            -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def interpolate(self, method: str) \
            -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def normalize(self, method: str) \
            -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def round(self, decimals: int) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def sort(self, *args) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    # ==========================================================================
    # Analysis
    # ==========================================================================

    # Basic Statistics
    # ----------------

    @abstractmethod
    def min(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def max(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def mean(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def median(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def kurtosis(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def skewness(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def describe(self) -> DataFrame:
        raise NotImplementedError

    # Time Series Statistics
    # ----------------------

    @abstractmethod
    def start(self) -> Union[Timestamp, List[Timestamp]]:
        raise NotImplementedError

    @abstractmethod
    def end(self) -> Union[Timestamp, List[Timestamp]]:
        raise NotImplementedError

    @abstractmethod
    def boundaries(self) \
            -> Union[Tuple[Timestamp, Timestamp],
                     List[Tuple[Timestamp, Timestamp]]]:
        raise NotImplementedError

    @abstractmethod
    def frequency(self) -> Union[Optional[str], List[Optional[str]]]:
        raise NotImplementedError

    @abstractmethod
    def resolution(self) -> Union['TimeSeries', 'TimeSeriesDataset']:
        raise NotImplementedError

    @abstractmethod
    def duration(self) -> Union[Timedelta, List[Timedelta]]:
        raise NotImplementedError



import json
from os import path as os_path
import pandas as pd
from pandas import Series
from timeatlas.config.constants import *
from timeatlas import Metadata

def check_directory_structure(path: str):
    """
    Check if a directory given as dataset path has a structure corresponding
    to the timeatlas standard

    Args:
        path: path encoded in a string

    Returns:
        A Boolean representing the validity of the structure
    """

    is_data = os_path.exists("{}/{}.{}".format(path, TIME_SERIES_FILENAME, TIME_SERIES_EXT))
    is_meta = os_path.exists("{}/{}.{}".format(path, METADATA_FILENAME, METADATA_EXT))

    if is_data is True and is_meta is False:
        return "timeseries"
    elif is_data is True and is_meta is True:
        return "timeseries with metadata"
    else:
        return None


def csv_to_dataframe(path: str) -> Series:
    df = pd.read_csv(path)
    df = df.set_index(pd.to_datetime(df["index"]))
    df = df.drop(columns=["index"])
    return df


def json_to_metadata(path: str) -> Metadata:
    with open(path) as json_file:
        meta = json.load(json_file)
    return Metadata(meta)



from typing import List, Callable, Tuple, Union

from timeatlas.time_series import TimeSeries
from timeatlas.abstract import AbstractBaseDetector, AbstractBaseModel
from timeatlas.processors import scaler, miscellaneous
from timeatlas.config.constants import TIME_SERIES_VALUES


class Surprise(AbstractBaseDetector):

    def __init__(self, model: AbstractBaseModel, error: Callable):
        super().__init__()

        # Params
        self.model = model
        self.error = error
        self.normalizer = None
        self._compute_thresholds_params = None

        # Intermediate results
        self.truth = None
        self.prediction = None
        self.surprise = None
        self.thresholds = None

        # Object state
        self._is_fitted = False

    def normalize(self, method: str):
        if method == "minmax":
            self.normalizer = scaler.minmax
        elif method == "zscore":
            self.normalizer = scaler.zscore
        return self

    def alerts(self, method: str = "quantile", thresholds: Union[float, List] = [0.85, 0.95]):
        self._compute_thresholds_params = (method, thresholds)
        return self

    def fit(self, ts: TimeSeries):

        # Set the default alerts if not custom
        if self._compute_thresholds_params is None:
            self.alerts()

        # Set the truth in the Surprise anomaly detector
        self.truth = ts

        # Predict and compute the actual surprise
        self.prediction, self.surprise = self.__compute_surprise(
            self.model, ts, self.error)

        # Normalize if desired
        if self.normalizer is not None:
            self.surprise = self.normalizer(self.surprise)

        # Compute the thresholds according to the given truth
        self.thresholds = self.__compute_thresholds(
            self._compute_thresholds_params[0],
            self._compute_thresholds_params[1],
            self.surprise
        )

        self._is_fitted = True
        return self

    def detect(self, ts: TimeSeries) -> TimeSeries:
        prediction, surprise = self.__compute_surprise(self.model, ts,
                                                       self.error)

        if self.normalizer is not None:
            surprise = self.normalizer(surprise)

        alerts = miscellaneous.ceil(surprise, self.thresholds)
        return alerts

    @staticmethod
    def __compute_surprise(m: AbstractBaseModel, ts: TimeSeries,
                           error_func: Callable) \
            -> Tuple['TimeSeries', 'TimeSeries']:
        prediction = m.predict(ts)
        surprise = ts.apply(error_func, prediction)
        return prediction, surprise

    @staticmethod
    def __compute_thresholds(method: str, thresholds: List,
                             ts: TimeSeries = None):
        res = []
        if method == "quantile":
            for k, v in enumerate(thresholds):
                res.append(ts.series[TIME_SERIES_VALUES].quantile(q=v))
        elif method == "threshold":
            res = thresholds
        else:
            raise Exception("method {} isn't implemented".format(method))
        return res




try:
    from torch.utils.data import Dataset
except ModuleNotFoundError:
    raise ModuleNotFoundError("Pytorch not found. Install with pip install torch")


class BaseDataset(Dataset):
    """
    Base Class for the PyTorch DataSets
    """

    def __init__(self, tsd):
        self.tsd = tsd



from typing import NoReturn, Union, Optional

from pandas import DataFrame
try:
    import fbprophet as fbp
except ModuleNotFoundError:
    raise ModuleNotFoundError("Could not find fbprophet. Install with: pip install fbprophet")

from timeatlas.abstract import AbstractBaseModel
from timeatlas.config.constants import (
    TIME_SERIES_VALUES,
    TIME_SERIES_CI_UPPER,
    TIME_SERIES_CI_LOWER,
    MODEL_TYPE_UNIVARIATE,
    MODEL_TYPE_MULTIVARIATE
)
from timeatlas.time_series import TimeSeries
from timeatlas.time_series_dataset import TimeSeriesDataset


class Prophet(AbstractBaseModel):

    def __init__(self):
        super().__init__()
        self.model = fbp.Prophet()
        self.y = None

    def fit(self, ts: Union[TimeSeries, TimeSeriesDataset],
            y: Optional[int] = None) -> NoReturn:
        """
        Fit a Prophet model to a time series. If given a TimeSeriesDataset, the
        optional argument y must be given to indicate on which component the
        model should be fitted to.

        Args:
            ts: TimeSeries or TimeSeriesDataset to fit
            y: Optional int of the component index in a TimeSeriesDataset

        Returns:
            NoReturn
        """
        super().fit(ts)

        # Prepare the model
        if isinstance(ts, TimeSeries):
            self.type = MODEL_TYPE_UNIVARIATE
            df = self.__prepare_time_series_for_prophet(self.X_train)

        elif isinstance(ts, TimeSeriesDataset):
            assert y is not None, "For multivariate prediction, the y " \
                                  "argument must be given."
            self.type = MODEL_TYPE_MULTIVARIATE
            self.y = y
            df = self.__prepare_time_series_dataset_for_prophet(
                self.X_train,
                self.y
            )

            # Add all components except y as extra regressor
            regressors = df.columns.to_list()
            regressors.remove("y")
            regressors.remove("ds")
            for r in regressors:
                self.model.add_regressor(r)
        else:
            ValueError('The fit method  accepts only TimeSeries or '
                       'TimeSeriesDataset as argument')

        self.model.fit(df)

    def predict(self, horizon: Union[str, TimeSeries, TimeSeriesDataset],
            freq: str = None) \
            -> TimeSeries:
        super().predict(horizon)

        # Prepare the data
        if self.type == MODEL_TYPE_UNIVARIATE:
            if isinstance(horizon, str):
                future = self.make_future_dataframe(horizon, freq)
                metadata = None
            elif isinstance(horizon, TimeSeries):
                future = self.__prepare_time_series_for_prophet(horizon.empty())
                metadata = horizon.metadata

        elif self.type == MODEL_TYPE_MULTIVARIATE:
            if isinstance(horizon, TimeSeriesDataset):
                horizon[:, self.y] = horizon[:, self.y].empty()
                future = self.__prepare_time_series_dataset_for_prophet(
                    horizon, self.y)
                metadata = horizon[:, self.y].data[self.y].metadata

        else:
            ValueError("horizon argument type isn't recognized")

        # Predict
        forecast = self.model.predict(future)
        forecast.rename(columns={"yhat": TIME_SERIES_VALUES,
                                 "yhat_lower": TIME_SERIES_CI_LOWER,
                                 "yhat_upper": TIME_SERIES_CI_UPPER},
                        inplace=True)
        df = forecast[[TIME_SERIES_VALUES,
                       TIME_SERIES_CI_LOWER,
                       TIME_SERIES_CI_UPPER]]
        df.index = forecast["ds"]

        # Register the prediction plot
        ts = TimeSeries(df, metadata)
        return ts

    @staticmethod
    def __prepare_time_series_for_prophet(ts: TimeSeries):
        df = ts.to_df().copy()
        df["ds"] = df.index
        df = df.reset_index(drop=True)
        df = df.rename(columns={"values": "y"})
        df.columns = df.columns.astype(str)
        return df

    @staticmethod
    def __prepare_time_series_dataset_for_prophet(tsd: TimeSeriesDataset,
            y: int):
        df = tsd.to_df().copy()
        df["ds"] = df.index
        df = df.reset_index(drop=True)
        df = df.rename(columns={y: "y"})
        df.columns = df.columns.astype(str)
        return df

    def make_future_dataframe(self, horizon: str, freq: str = None):
        index = self.make_future_index(horizon, freq)
        df = DataFrame(data=index.to_series(), columns=["ds"])
        df = df.reset_index(drop=True)
        return df



from typing import NoReturn
import pandas as pd


class AnomalySetLabeler:
    def __init__(self):
        """

        Class of the AnomalyGenerator Labels

        """
        self.annotation = {}

    def create_operation_dict(self, coordinates: list, param: str, function_name: str, name: str,
            outfile: str) -> NoReturn:
        """

        Create a dictionary of all the newly added anomalies.

        Args:
            coordinates: index of the start and end of the anomaly
            param: parameters used to create the anomaly
            function_name: name of the function that created the anomaly
            name: name of the dataframe used to insert the anomalies into. Filename
            outfile: Filename to save the labels to #deprecated

        Returns: NoReturn

        """
        coordinates = sorted(coordinates, key=lambda x: x[0])
        # TODO: This offset style is not good since it is redundant with he code in utils.insert
        offset = 0
        for coords in coordinates:
            try:
                next_key = int(max(self.annotation.keys())) + 1
            except:
                next_key = 0
            self.annotation[next_key] = {'start': coords[0] + offset,
                                         'end': coords[1] + offset,
                                         'function_name': function_name,
                                         'parameters': param,
                                         'dataframe_name': f'{outfile}_data/{name}/data.csv'
                                         }

            if param == 'insert':
                offset += coords[1] - coords[0]

    def finalize(self):
        self.annotation = pd.DataFrame.from_dict({i: self.annotation[i] for i in self.annotation.keys()},
                                                 orient='index')




class Sensor:
    """ Defines a sensor """

    def __init__(self, id: int, name: str):
        self.id = id
        self.name = name

    def __repr__(self):
        return "Sensor ID {}".format(self.id)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, NoReturn, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from timeatlas.time_series import TimeSeries
    from timeatlas.time_series_dataset import TimeSeriesDataset

from pandas import Timedelta, date_range, infer_freq


class AbstractBaseModel(ABC):
    """ Abstract class to define methods to implement
    for a Model class inspired by Scikit Learn API.
    """

    def __init__(self):
        self._is_fitted = False
        self.X_train = None
        self.type = None

    @abstractmethod
    def fit(self, series: Union["TimeSeries", "TimeSeriesDataset"]) \
            -> NoReturn:
        """ Fit a model """
        self._is_fitted = True
        self.X_train = series

    @abstractmethod
    def predict(self, horizon: Union[str, "TimeSeries", "TimeSeriesDataset"]) \
            -> Any:
        """ Usage of the model to predict values """
        if self._is_fitted is False:
            raise Exception('fit() must be called before predict()')

    def make_future_index(self, horizon: str, freq: str = None):
        """ Creates a DatetimeIndex from the last timestamp given
        in the training set X_train for a given horizon.

        Args:
            horizon: str as in https://pandas.pydata.org/pandas-docs/stable/user_guide/timedeltas.html
            freq: frequency in DateOffset string https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

        Returns:
            DatetimeIndex
        """
        start = self.X_train.series.index[-1]
        end = self.X_train.series.index[-1] + Timedelta(horizon)
        if freq is not None:
            freq = freq
        else:
            freq = infer_freq(self.X_train.series.index)
        return date_range(start=start, end=end, freq=freq)[0:]



import setuptools
import subprocess
from pathlib import Path

with open("README.md", "r") as fh:
    long_description = fh.read()


def read_requirements(path):
    return list(Path(path).read_text().splitlines())

base_req = read_requirements('requirements/src_base.txt')
torch_req = read_requirements('requirements/src_torch.txt')
prophet_req = read_requirements('requirements/src_prophet.txt')
all_req = base_req + prophet_req + torch_req


def get_branch():
    try:
        return subprocess.check_output(['git', 'branch', '--show-current']) \
            .decode('ascii').strip()
    except Exception:
        return None


def get_commit():
    try:
        return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']) \
            .decode('ascii').strip()
    except Exception:
        return 'unknown'


def get_commit_time(commit):
    try:
        return subprocess.check_output(
            ['git', 'show', '-s', '--format="%ct"', commit]) \
                   .decode('ascii').strip()[1:-1]
    except Exception:
        return 'unknown'


def create_version():
    version = open('version.txt', 'r').read().strip()
    branch = get_branch()
    if branch == "develop" or branch is None:
        commit_time = get_commit_time(get_commit())
        version = "{}{}{}".format(version, '.dev', commit_time)
    elif branch == "master":
        version
    return version


setuptools.setup(
    name="timeatlas",
    version=create_version(),
    author="Frédéric Montet",
    author_email="frederic.montet@hefr.ch",
    description="A time series data manipulation tool for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=setuptools.find_packages('src', exclude='tests'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=base_req,
    extras_require={'all': all_req,
                   'torch': torch_req,
                   'prophet': prophet_req
                   },
)


from pandas import Series


class Scaler:

    @staticmethod
    def minmax(s: Series) -> Series:
        scaled_series = (s - s.min()) / (s.max() - s.min())
        return scaled_series

    @staticmethod
    def zscore(s: Series) -> Series:
        scaled_series = (s - s.mean()) / s.std()
        return scaled_series



import os
from typing import NoReturn, Any
import pickle


def ensure_dir(file_path: str) -> NoReturn:
    """
    Make sure that a path to a directory exists, or creates its directory leafs

    Args:
        file_path: the path to the file/directory
    """
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)


def to_pickle(obj: Any, path: str) -> NoReturn:
    """
    Export a object in Pickle on your file system

    Args:
        obj: Object to serialize
        path: str of the path to the target directory
    """
    ensure_dir(path)
    with open(path, 'wb') as file:
        pickle.dump(obj, file, pickle.HIGHEST_PROTOCOL)



import numpy as np

from .base_dataset import BaseDataset
from timeatlas.time_series_dataset import TimeSeriesDataset


class TimeSeriesClassificationDataset(BaseDataset):
    """
    A DataLoader for the classification of complete TimeSeries, where X: TimeSeries and y: label of the TimeSeries
    """

    def __init__(self, timeseriesdataset: TimeSeriesDataset):
        super(TimeSeriesClassificationDataset, self).__init__(tsd=timeseriesdataset)
        self.data = np.array([ts.series for ts in timeseriesdataset])
        self.labels = [ts.label for ts in timeseriesdataset]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item], self.labels[item]
